﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Controls.Primitives;
using System.Windows.Input;
using System.Windows.Media;
using UniversalEditor.Base;

namespace TextEditor
{
	public delegate void KeyDownDelegate(KeyEventArgs e);

	public class UfeTextBox : Grid
	{
		public event KeyDownDelegate OnKeyDownEvent
		{
			add { _textBox.OnKeyDownEvent += value; }
			remove { _textBox.OnKeyDownEvent -= value; }
		}

		private readonly ScrollBar _verticalScroll;
		private readonly ScrollBar _horizontalScroll;
		private readonly FindLineCtrl _findLineCtrl;
		private readonly  FindWordCtrl _findWordCtrl;
		private readonly  DocumentChangedCtrl _documentChangedCtrl;
		private readonly  UfeTextBoxInner _textBox;
		private readonly UfePopup _popup;

		private readonly MenuItem _readonlyMenuItem = new MenuItem();
		private readonly ReadonlyCtrl _readonlyCtrl;
		private readonly DocumentDeletedCtrl _documentDeletedCtrl;

		public UfeTextBoxInner TextBox
		{
			get { return _textBox; }
		}
		
		public UfeTextBox(IEditorOwner owner, UfeTextDocument doc)
		{
			BeginInit();

			// background
			UserControl background = new UserControl();
			background.Background = Brushes.Transparent;
			SetColumnSpan(background, 2);
			Children.Add(background);

			// scroll
			_verticalScroll = new ScrollBar();
			_verticalScroll.Width = 20;
			_verticalScroll.Visibility = Visibility.Collapsed;
			_verticalScroll.Orientation = Orientation.Vertical;
			_verticalScroll.Margin = new Thickness(0);
			_verticalScroll.Padding = new Thickness(0);

			SetColumn(_verticalScroll, 2);
			SetRow(_verticalScroll, 0);
			SetRowSpan(_verticalScroll, 3);
			Children.Add(_verticalScroll);

			_horizontalScroll = new ScrollBar();
			_horizontalScroll.Height = 20;
			_horizontalScroll.Visibility = Visibility.Collapsed;
			_horizontalScroll.Orientation = Orientation.Horizontal;
			_horizontalScroll.Margin = new Thickness(0);
			_horizontalScroll.Padding = new Thickness(0);

			SetColumnSpan(_horizontalScroll, 2);
			SetRow(_horizontalScroll, 3);
			Children.Add(_horizontalScroll);
			
			// find line
			_findLineCtrl = new FindLineCtrl(this);
			_findLineCtrl.Visibility = Visibility.Collapsed;
			_findLineCtrl.Margin = new Thickness(0);
			_findLineCtrl.Padding = new Thickness(0);
			SetColumn(_findLineCtrl, 1);
			SetRow(_findLineCtrl, 0);

			// find word
			_findWordCtrl = new FindWordCtrl(this);
			_findWordCtrl.Visibility = Visibility.Collapsed;
			_findWordCtrl.Margin = new Thickness(0);
			_findWordCtrl.Padding = new Thickness(0);
			SetColumn(_findWordCtrl, 1);
			SetRow(_findWordCtrl, 0);

			// readonly
			_readonlyCtrl = new ReadonlyCtrl(this);
			_readonlyCtrl.Visibility = Visibility.Collapsed;
			_readonlyCtrl.HorizontalAlignment = HorizontalAlignment.Center;
			_readonlyCtrl.Margin = new Thickness(0);
			_readonlyCtrl.Padding = new Thickness(0);
			SetColumnSpan(_readonlyCtrl, 2);
			SetRow(_readonlyCtrl, 2);

			// popup
			_popup = new UfePopup(this);
			_popup.Visibility = Visibility.Collapsed;
			_popup.Placement = PlacementMode.Custom;
			_popup.CustomPopupPlacementCallback = CustomPopupPlacementCallback;

			// file changed
			_documentChangedCtrl = new DocumentChangedCtrl(this);
			_documentChangedCtrl.Visibility = Visibility.Collapsed;
			_documentChangedCtrl.HorizontalAlignment = HorizontalAlignment.Center;
			_documentChangedCtrl.Margin = new Thickness(0);
			_documentChangedCtrl.Padding = new Thickness(0);
			SetColumnSpan(_documentChangedCtrl, 2);
			SetRow(_documentChangedCtrl, 2);

			// file changed
			_documentDeletedCtrl = new DocumentDeletedCtrl(owner, this);
			_documentDeletedCtrl.Visibility = Visibility.Collapsed;
			_documentDeletedCtrl.HorizontalAlignment = HorizontalAlignment.Center;
			_documentDeletedCtrl.Margin = new Thickness(0);
			_documentDeletedCtrl.Padding = new Thickness(0);
			SetColumnSpan(_documentDeletedCtrl, 2);
			SetRow(_documentDeletedCtrl, 2);

			// _textBox
			_textBox = new UfeTextBoxInner(doc, _verticalScroll, _horizontalScroll, _findLineCtrl, _findWordCtrl, _documentChangedCtrl, _documentDeletedCtrl, _readonlyCtrl, _popup);
			_textBox.ClipToBounds = true;
			_textBox.Cursor = Cursors.IBeam;
			SetRowSpan(_textBox, 3);
			SetColumnSpan(_textBox, 2);

			Children.Add(_textBox);
			Children.Add(_findLineCtrl);
			Children.Add(_findWordCtrl);
			Children.Add(_readonlyCtrl);
			Children.Add(_popup);
			Children.Add(_documentChangedCtrl);
			Children.Add(_documentDeletedCtrl);

			ColumnDefinitions.Add(new ColumnDefinition { Width = new GridLength(100, GridUnitType.Star) });
			ColumnDefinitions.Add(new ColumnDefinition { Width = new GridLength(1, GridUnitType.Auto) });
			ColumnDefinitions.Add(new ColumnDefinition { Width = new GridLength(1, GridUnitType.Auto) });

			RowDefinitions.Add(new RowDefinition { Height = new GridLength(1, GridUnitType.Auto) });
			RowDefinitions.Add(new RowDefinition { Height = new GridLength(100, GridUnitType.Star) });
			RowDefinitions.Add(new RowDefinition { Height = new GridLength(1, GridUnitType.Auto) });
			RowDefinitions.Add(new RowDefinition { Height = new GridLength(1, GridUnitType.Auto) });

			ClipToBounds = true;
			Margin = new Thickness(0);
			_popup.PlacementTarget = _textBox;

			// initialize menu
			_readonlyMenuItem.Header = "Readonly";
			_readonlyMenuItem.IsCheckable = true;
			_readonlyMenuItem.IsChecked = IsReadonly;
			_readonlyMenuItem.Checked += (sender, args) => IsReadonly = true;
			_readonlyMenuItem.Unchecked += (sender, args) => IsReadonly = false;

			if (doc.IsNewFile)
				IsReadonly = false;

			_textBox.TabIndex = 1;
			_textBox.Focus();

			EndInit();
		}

		private CustomPopupPlacement[] CustomPopupPlacementCallback(Size popupSize, Size targetSize, Point offset)
		{
			CustomPopupPlacement placement1 = new CustomPopupPlacement(_textBox.GetCursorPosition(), PopupPrimaryAxis.Vertical);
			CustomPopupPlacement placement2 = new CustomPopupPlacement(_textBox.GetCursorPosition(), PopupPrimaryAxis.Horizontal);

			CustomPopupPlacement[] ttplaces = new[] { placement1, placement2 };
			return ttplaces;
		}

		public void ShowLine(int number)
		{
			_textBox.ShowLine(number);
		}

		public UfeTextDocument Document
		{
			get { return _textBox.Document; }
		}

		public void FindText(string text, bool useRegister)
		{
			_textBox.FindText(text, useRegister);
		}

		public void HideAllFindControls()
		{
			_textBox.SetTopPanelVisibility(UfeTextBoxInner.TopPanelVisibility.None);
		}

		public void HideAllBottomControls()
		{
			_textBox.SetBottomPanelVisibility(UfeTextBoxInner.BottomPanelVisibility.None);
		}

		public bool IsReadonly
		{
			get { return _textBox.IsReadonly; }
			set
			{
				if (_textBox.IsReadonly == value)
					return;

				_textBox.IsReadonly = value;
				_readonlyMenuItem.IsChecked = value;

				HideAllBottomControls();
			}
		}

		public void CallOnRender()
		{
			_textBox.InvalidateVisual();
		}

		public List<MenuItem> GenerateBasicMenu()
		{
			List<MenuItem> menu = new List<MenuItem>();
			menu.Add(_readonlyMenuItem);

			return menu;
		}

		public void Insert(HintItem word)
		{
			_textBox.Insert(word);
		}
	}
}
