using System;
using System.Collections.Generic;
using System.Globalization;
using System.IO;
using System.Linq;
using System.Text;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Controls.Primitives;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using UniversalEditor.Base.Options;

namespace TextEditor
{
	public class UfeTextBoxInner : UserControl
	{
		internal enum TopPanelVisibility { None, Line, Word }
		internal enum BottomPanelVisibility { None, Readonly, Changed, Deleted }

		public event KeyDownDelegate OnKeyDownEvent;

		private readonly double fontSize = Options.Instance.FontSize;
		private double _lineStartOffset = 30;
		public readonly static char[] CtrlSplitChars = new[] { ' ', ')', '(', '/', '\\', '\t', '.', ',', '"', '\'', ':', ';', '<', '>', '|' };
		internal static readonly Typeface Typeface = new Typeface(new FontFamily(Options.Instance.FontName), Application.Current.MainWindow.FontStyle, Application.Current.MainWindow.FontWeight, Application.Current.MainWindow.FontStretch);
		
		private readonly UfeTextDocument _document;
		private readonly ScrollBar _verticalScroll;
		private readonly ScrollBar _horizontalScroll;
		private readonly FindLineCtrl _findLineCtrl;
		private readonly FindWordCtrl _findWordCtrl;
		private readonly DocumentChangedCtrl _documentChangedCtrl;
		private readonly DocumentDeletedCtrl _documentDeletedCtrl;
		private readonly ReadonlyCtrl _readonlyCtrl;
		private readonly UfePopup _popup;

		private readonly List<TextPosition> _linesGeometry = new List<TextPosition>();
		
		private int _renderedStartLineIndex = 0;
		private double _renderedStartLineOffset = 0;
		private int _renderedLineCount = 0;
		private DateTime _lastMouseDownEvent = DateTime.MinValue;
		private CharPosition _lastMouseDownPosition;
		private bool _isLastMouseDownDouble = false;

		private Point _cursorPosition = new Point();

		private UfeTextLineBase _longestLine;
		private double _longestLineWidth = -1;

		private CharPosition _startSelection = new CharPosition(0, 0);
		private CharPosition _endSelection = new CharPosition(0, 0);
		private bool _isReadonly = false;

		public UfeTextDocument Document
		{
			get { return _document; }
		}

		public bool IsReadonly
		{
			get { return _isReadonly; }
			set
			{
				if (_isReadonly == value)
					return;

				_isReadonly = value;
				InvalidateVisual();
			}
		}

		public UfeTextBoxInner(UfeTextDocument document, ScrollBar verticalScroll, ScrollBar horizontalScroll, FindLineCtrl findLineCtrl, FindWordCtrl findWordCtrl, DocumentChangedCtrl documentChangedCtrl, DocumentDeletedCtrl documentDeletedCtrl, ReadonlyCtrl readonlyCtrl, UfePopup popup)
		{
			if (document == null) 
				throw new ArgumentNullException("document");
			if (verticalScroll == null)
				throw new ArgumentNullException("verticalScroll");
			if (horizontalScroll == null)
				throw new ArgumentNullException("horizontalScroll");
			if (findLineCtrl == null)
				throw new ArgumentNullException("findLineCtrl");
			if (findWordCtrl == null)
				throw new ArgumentNullException("findWordCtrl");
			if (documentChangedCtrl == null)
				throw new ArgumentNullException("documentChangedCtrl");
			if (documentDeletedCtrl == null) 
				throw new ArgumentNullException("documentDeletedCtrl");
			if (readonlyCtrl == null) 
				throw new ArgumentNullException("readonlyCtrl");
			if (popup == null)
				throw new ArgumentNullException("popup");

			_document = document;
			_verticalScroll = verticalScroll;
			_horizontalScroll = horizontalScroll;
			_findLineCtrl = findLineCtrl;
			_findWordCtrl = findWordCtrl;
			_documentChangedCtrl = documentChangedCtrl;
			_documentDeletedCtrl = documentDeletedCtrl;
			_readonlyCtrl = readonlyCtrl;
			_popup = popup;

			_verticalScroll.ValueChanged += _verticalScroll_ValueChanged;
			_verticalScroll.MouseLeftButtonDown += VerticalScrollOnMouseLeftButtonDown;
			_verticalScroll.MouseLeftButtonUp += VerticalScrollOnMouseLeftButtonDown;
			_verticalScroll.MouseMove += _verticalScroll_MouseMove;

			_horizontalScroll.ValueChanged += _horizontalScroll_ValueChanged;
			_horizontalScroll.MouseLeftButtonDown += HorizontalScrollOnMouseLeftButtonDown;
			_horizontalScroll.MouseLeftButtonUp += HorizontalScrollOnMouseLeftButtonDown;
			_horizontalScroll.MouseMove += _horizontalScroll_MouseMove;

			_document.ContentChanged += OnContentChanged;
			_document.FileDeleted += OnFileDeleted;

			FindTheLongestLine();

			Background = Brushes.Transparent;
			Focusable = true;
			Keyboard.Focus(this);

			if (IsReadonly)
				SetBottomPanelVisibility(BottomPanelVisibility.Readonly);

			FocusVisualStyle = null;
		}

		private void FindTheLongestLine()
		{
			for (int index = 0; index < _document.LineCount; index++)
			{
				UfeTextLineBase line = _document.GetLine(index);
				double lineWidth = line.GetWidthInPixels();

				if (_longestLineWidth < lineWidth)
				{
					_longestLine = line;
					_longestLineWidth = lineWidth;
				}
			}
		}

		private void OnFileDeleted()
		{
			if (!Dispatcher.CheckAccess())
			{
				Dispatcher.Invoke(OnFileDeleted);
				return;
			}

			SetBottomPanelVisibility(BottomPanelVisibility.Deleted);
		}

		private void OnContentChanged()
		{
			if (!Dispatcher.CheckAccess())
			{
				Dispatcher.Invoke(OnContentChanged);
				return;
			}

			if (_isReadonly)
			{
				_document.Reload();
				InvalidateVisual();
			}
			else
			{
				SetBottomPanelVisibility(BottomPanelVisibility.Changed);
			}
		}

		void _verticalScroll_MouseMove(object sender, MouseEventArgs e)
		{
			e.Handled = true;
		}

		void _horizontalScroll_MouseMove(object sender, MouseEventArgs e)
		{
			e.Handled = true;
		}

		private void VerticalScrollOnMouseLeftButtonDown(object sender, MouseButtonEventArgs mouseButtonEventArgs)
		{
			mouseButtonEventArgs.Handled = true;
		}

		private void HorizontalScrollOnMouseLeftButtonDown(object sender, MouseButtonEventArgs mouseButtonEventArgs)
		{
			mouseButtonEventArgs.Handled = true;
		}

		private void _verticalScroll_ValueChanged(object sender, RoutedPropertyChangedEventArgs<double> e)
		{
			_renderedStartLineIndex = Convert.ToInt32(_verticalScroll.Value);
			InvalidateVisual();
		}

		private void _horizontalScroll_ValueChanged(object sender, RoutedPropertyChangedEventArgs<double> e)
		{
			_renderedStartLineOffset = Convert.ToInt32(_horizontalScroll.Value);
			InvalidateVisual();
		}

		internal static string ReplaceTabs(string text)
		{
			return ReplaceTabs(text, 0);
		}
		
		internal static string ReplaceTabs(string text, int beginOffset)
		{
			StringBuilder builder = new StringBuilder();

			for (int index = 0; index < text.Length; index++)
			{
				char ch = text[index];

				if (ch == '\t')
				{
					int tabCount = 4 - beginOffset % 4;

					for (int i = 0; i < tabCount; i++)
						builder.Append(' ');

					beginOffset += tabCount;
				}
				else
				{
					builder.Append(ch);
					beginOffset++;
				}
			}

			return builder.ToString();
		}

		internal void SetTopPanelVisibility(TopPanelVisibility visibility)
		{
			switch (visibility)
			{
				case TopPanelVisibility.Line:
					_findWordCtrl.Visibility = Visibility.Collapsed;
					_findLineCtrl.Visibility = Visibility.Visible;

					_findLineCtrl.txtNumber.UpdateLayout();
					_findLineCtrl.txtNumber.Focus();
					_findLineCtrl.txtNumber.SelectAll();
					break;
				case TopPanelVisibility.Word:
					_findLineCtrl.Visibility = Visibility.Collapsed;
					_findWordCtrl.Visibility = Visibility.Visible;

					_findWordCtrl.txtWord.UpdateLayout();
					_findWordCtrl.txtWord.Focus();
					_findWordCtrl.txtWord.SelectAll();
					break;
				case TopPanelVisibility.None:
					_findLineCtrl.Visibility = Visibility.Collapsed;
					_findWordCtrl.Visibility = Visibility.Collapsed;

					Keyboard.Focus(this);
					break;
			}
		}

		internal void SetBottomPanelVisibility(BottomPanelVisibility visibility)
		{
			switch (visibility)
			{
				case BottomPanelVisibility.Deleted:
					_readonlyCtrl.Visibility = Visibility.Collapsed;
					_documentChangedCtrl.Visibility = Visibility.Collapsed;
					_documentDeletedCtrl.Visibility = Visibility.Visible;
					Keyboard.Focus(this);
					break;
				case BottomPanelVisibility.Changed:
					_readonlyCtrl.Visibility = Visibility.Collapsed;
					_documentChangedCtrl.Visibility = Visibility.Visible;
					_documentDeletedCtrl.Visibility = Visibility.Collapsed;
					Keyboard.Focus(this);
					break;
				case BottomPanelVisibility.Readonly:
					_readonlyCtrl.Visibility = Visibility.Visible;
					_documentChangedCtrl.Visibility = Visibility.Collapsed;
					_documentDeletedCtrl.Visibility = Visibility.Collapsed;
					Keyboard.Focus(this);
					break;
				case BottomPanelVisibility.None:
					_readonlyCtrl.Visibility = Visibility.Collapsed;
					_documentChangedCtrl.Visibility = Visibility.Collapsed;
					_documentDeletedCtrl.Visibility = Visibility.Collapsed;
					Keyboard.Focus(this);
					break;
			}
		}

		protected override void OnRender(DrawingContext drawingContext)
		{
			base.OnRender(drawingContext);
			
			_linesGeometry.Clear();

			double renderHeight = RenderSize.Height;
			
			double linePosition = 0;
			_renderedLineCount = 0;

			FormattedText testNumberText = new FormattedText(((_renderedStartLineIndex + 1) * 10).ToString(), CultureInfo.CurrentUICulture, FlowDirection.LeftToRight, Typeface, fontSize, Application.Current.MainWindow.Foreground);

			double areaNumberWidth = _lineStartOffset = testNumberText.Width;
			_lineStartOffset += 10 - _renderedStartLineOffset;
			
			for (int index = _renderedStartLineIndex; index < _document.LineCount && linePosition < renderHeight; index++, _renderedLineCount++)
			{
				UfeTextLineBase text = _document.GetLine(index);

				if (_isReadonly && !text.IsRendarableInReadonly())
					continue;
				
				FormattedText formattedText = new FormattedText(ReplaceTabs(text.TextAsString), CultureInfo.CurrentUICulture, FlowDirection.LeftToRight, Typeface, fontSize, Application.Current.MainWindow.Foreground) { Trimming = TextTrimming.None };

				if (text.Length == 0)
				{
					_linesGeometry.Add(new TextPosition(text.TextAsString, formattedText, linePosition, linePosition + fontSize, index));

					if (index == _startSelection.Row && _startSelection == _endSelection)
					{
						_cursorPosition = new Point(_lineStartOffset, linePosition);
						drawingContext.DrawLine(new Pen(Brushes.Black, 1), new Point(_lineStartOffset, linePosition), new Point(_lineStartOffset, linePosition + fontSize));
					}

					linePosition += fontSize;
				}
				else
				{
					_linesGeometry.Add(new TextPosition(text.TextAsString, formattedText, linePosition, linePosition + formattedText.Height, index));

					long startRow = Math.Min(_startSelection.Row, _endSelection.Row);
					long endRow = Math.Max(_startSelection.Row, _endSelection.Row);

					int startCol =
						_startSelection.Row < _endSelection.Row ? _startSelection.Col :
						_startSelection.Row > _endSelection.Row ? _endSelection.Col :
						Math.Min(_startSelection.Col, _endSelection.Col);

					int endCol =
						_startSelection.Row > _endSelection.Row ? _startSelection.Col :
						_startSelection.Row < _endSelection.Row ? _endSelection.Col :
						Math.Max(_startSelection.Col, _endSelection.Col);

					int startSel =
						startRow < index && endRow >= index ? 0 :
						startRow == index ? Math.Min(startCol, text.Length) : -1;
					int endSel =
						endRow < index ? -1 :
						endRow == index ? Math.Min(endCol, text.Length) :
						endRow > index && startRow <= index ? text.Length : -1;

					DrawLine(drawingContext, text, startSel, endSel, linePosition);

					linePosition += formattedText.Height;
				}
			}

			// draw line number
			drawingContext.DrawRectangle(Brushes.LightGray, null, new Rect(0, 0, areaNumberWidth, renderHeight));

			for (int index = 0; index < _linesGeometry.Count; index++)
			{
				TextPosition textPosition = _linesGeometry[index];

				FormattedText lineNumber = new FormattedText((textPosition.Line + 1).ToString(), CultureInfo.CurrentUICulture, FlowDirection.LeftToRight, Typeface, fontSize, Application.Current.MainWindow.Foreground) { Trimming = TextTrimming.None, TextAlignment = TextAlignment.Center };
				drawingContext.DrawText(lineNumber, new Point(areaNumberWidth / 2.0, textPosition.Top));
			}
			//

			if (_renderedLineCount != _document.LineCount)
			{
				_verticalScroll.Maximum = _document.LineCount - _renderedLineCount;
				_verticalScroll.Value = _renderedStartLineIndex;
				_verticalScroll.ViewportSize = _renderedLineCount;
				_verticalScroll.Visibility = Visibility.Visible;
			}
			else
			{
				_verticalScroll.Visibility = Visibility.Collapsed;
			}

			double neededWidth = _longestLineWidth + areaNumberWidth + 10;
			if (neededWidth > RenderSize.Width)
			{
				_horizontalScroll.Maximum = neededWidth - RenderSize.Width;
				_horizontalScroll.Value = _renderedStartLineOffset;
				_horizontalScroll.ViewportSize = RenderSize.Width;
				_horizontalScroll.Visibility = Visibility.Visible;
			}
			else
			{
				_horizontalScroll.Visibility = Visibility.Collapsed;
			}
		}

		private void DrawLine(DrawingContext drawingContext, UfeTextLineBase line, int startSel, int endSel, double linePosition)
		{
			double x = _lineStartOffset;
			int chCount = 0;

			if (startSel == -1 && endSel == -1)
			{
				foreach (UfeTextStyle ufeTextStyle in line.Text)
				{
					string text = ReplaceTabs(ufeTextStyle.Text.ToString(), chCount);
					FormattedText part1 = new FormattedText(text, CultureInfo.CurrentUICulture, FlowDirection.LeftToRight, Typeface, fontSize, ufeTextStyle.Brush) { Trimming = TextTrimming.CharacterEllipsis, MaxLineCount = 1 };
					drawingContext.DrawText(part1, new Point(x, linePosition));

					x += part1.WidthIncludingTrailingWhitespace;
					chCount += text.Length;
				}
			}

			if (startSel != -1)
			{
				foreach (UfeTextStyle ufeTextStyle in line.Substring(0, startSel))
				{
					string text = ReplaceTabs(ufeTextStyle.Text.ToString(), chCount);
					FormattedText part1 = new FormattedText(text, CultureInfo.CurrentUICulture, FlowDirection.LeftToRight, Typeface, fontSize, ufeTextStyle.Brush) { Trimming = TextTrimming.CharacterEllipsis, MaxLineCount = 1 };
					drawingContext.DrawText(part1, new Point(x, linePosition));

					x += part1.WidthIncludingTrailingWhitespace;
					chCount += text.Length;
				}
			}

			if (startSel != -1 && endSel != -1)
			{
				if (startSel == endSel)
				{
					_cursorPosition = new Point(x, linePosition);
					drawingContext.DrawLine(new Pen(Brushes.Black, 1), new Point(x, linePosition), new Point(x, linePosition + fontSize));
				}
				else
				{
					foreach (UfeTextStyle ufeTextStyle in line.Substring(startSel, endSel - startSel))
					{
						string text = ReplaceTabs(ufeTextStyle.Text.ToString(), chCount);
						FormattedText part2 = new FormattedText(text, CultureInfo.CurrentUICulture, FlowDirection.LeftToRight, Typeface, fontSize, ufeTextStyle.Brush) { Trimming = TextTrimming.CharacterEllipsis, MaxLineCount = 1 };
						drawingContext.DrawGeometry(Brushes.LightSkyBlue, null, part2.BuildHighlightGeometry(new Point(x, linePosition)));
						drawingContext.DrawText(part2, new Point(x, linePosition));

						x += part2.WidthIncludingTrailingWhitespace;
						chCount += text.Length;
					}
				}
			}

			if (endSel != -1)
			{
				foreach (UfeTextStyle ufeTextStyle in line.Substring(endSel))
				{
					string text = ReplaceTabs(ufeTextStyle.Text.ToString(), chCount);
					FormattedText part3 = new FormattedText(text, CultureInfo.CurrentUICulture, FlowDirection.LeftToRight, Typeface, fontSize, ufeTextStyle.Brush) { Trimming = TextTrimming.CharacterEllipsis, MaxLineCount = 1 };
					drawingContext.DrawText(part3, new Point(x, linePosition));

					x += part3.WidthIncludingTrailingWhitespace;
					chCount += text.Length;
				}
			}
		}

		protected override void OnMouseWheel(MouseWheelEventArgs e)
		{
			_renderedStartLineIndex = Math.Min(_document.LineCount ,Math.Max(0, _renderedStartLineIndex - e.Delta / 40));
			InvalidateVisual();
			e.Handled = true;
		}

		internal Point GetCursorPosition()
		{
			return new Point(_cursorPosition.X, _cursorPosition.Y + fontSize);
		}

		private CharPosition GetChartPosition(Point location)
		{
			for (int index = 0; index < _linesGeometry.Count; index++)
			{
				TextPosition text = _linesGeometry[index];

				if (text == null || text.Top > location.Y || text.Buttom < location.Y)
					continue;
				
				if (string.IsNullOrEmpty(text.TextAsString.Trim()))
					return new CharPosition(text.TextAsString.Length, text.Line);

				GeometryGroup line = (GeometryGroup)text.Text.BuildGeometry(new Point(_lineStartOffset, text.Top));

//				if (line.FillContains(location))
				int finalCol = 0;
				for (int marker1 = 0; marker1 < line.Children.Count; marker1++)
				{
					GeometryGroup symbols = (GeometryGroup)line.Children[marker1];

					for (int marker2 = 0; marker2 < symbols.Children.Count; marker2++)
					{
						Geometry item = symbols.Children[marker2];
						Rect bounds = item.GetRenderBounds(null);

						if (location.X >= bounds.Left && location.X <= bounds.Right)
						{
							finalCol += marker2 + 1;
							for (int i = 0; i < finalCol; i++)
							{
								char ch = text.TextAsString[i];

								if (ch == ' ' || ch == '\t')
									finalCol++;
							}

							return new CharPosition(finalCol, text.Line);
						}
						if (location.X < bounds.Left)
						{
							finalCol += marker2 + 1;
							for (int i = 0; i < finalCol; i++)
							{
								char ch = text.TextAsString[i];

								if (ch == ' ' || ch == '\t')
									finalCol++;
							}

							return new CharPosition(finalCol - 1, text.Line);
						}
					}

					finalCol += symbols.Children.Count;
				}

				return new CharPosition(text.TextAsString.Length, text.Line);
			}

			return new CharPosition(_linesGeometry.Last().TextAsString.Length, _linesGeometry.Last().Line);
		}

		protected override void OnMouseLeftButtonDown(MouseButtonEventArgs e)
		{
			if (DateTime.UtcNow - _lastMouseDownEvent > TimeSpan.FromMilliseconds(300))
			{
				_isLastMouseDownDouble = false;

				if (!Keyboard.IsKeyDown(Key.LeftShift))
				{
					Point location = e.GetPosition(this);
					CharPosition temp = GetChartPosition(location);

					if (temp != null)
						SetPosition(temp);

					_lastMouseDownPosition = temp;
				}
			}
			else
			{
				_isLastMouseDownDouble = true;

				Point location = e.GetPosition(this);
				CharPosition temp = GetChartPosition(location);

				_lastMouseDownPosition = temp;

				if (temp != null)
				{
					UfeTextLineBase line = _document.GetLine(temp.Row);
					int start = Math.Max(line.TextAsString.LastIndexOfAny(CtrlSplitChars, Math.Min(temp.Col, line.TextAsString.Length - 1)) + 1, 0);
					int end = line.TextAsString.IndexOfAny(CtrlSplitChars, temp.Col);

					if (end < 0)
						end = line.TextAsString.Length;

					SetPosition(new CharPosition(start, temp.Row), new CharPosition(end, temp.Row));
				}
			}
			
			_lastMouseDownEvent = DateTime.UtcNow;

			e.Handled = true;
			Keyboard.Focus(this);
			CaptureMouse();
		}

		protected override void OnPreviewTextInput(TextCompositionEventArgs e)
		{
			OnTextInputEngine(e);
		}

		public void OnTextInputEngine(TextCompositionEventArgs e)
		{
			string text = e.Text;

			if (_isReadonly)
				return;

			if (string.IsNullOrEmpty(text))
				return;

			OnRemoveSelection();

			int newCol = _endSelection.Col;
			int newRow = _endSelection.Row;

			UfeTextLineBase line = _document.GetLine(newRow);
			line.Insert(newCol, text);

			newCol += text.Length;

			SetPosition(new CharPosition(newCol, newRow));

			ShowHint();
			_document.SetIsChangedToTrue();

			if (_longestLine == line)
				_longestLineWidth = _longestLine.GetWidthInPixels();

			e.Handled = true;
		}

		public void Insert(HintItem word)
		{
			UfeTextLineBase line = _document.GetLine(_endSelection.Row);

			line.Remove(_endSelection.Col - word.Prefix.Length, word.Prefix.Length);
			line.Insert(_endSelection.Col - word.Prefix.Length, word.Prefix);
			line.Insert(_endSelection.Col, word.Value);

			SetPosition(new CharPosition(_endSelection.Col + word.Value.Length, _endSelection.Row));

			if (_longestLine == line)
				_longestLineWidth = _longestLine.GetWidthInPixels();

			_document.SetIsChangedToTrue();
		}

		public void FindText(string text, bool useRegister)
		{
			StringComparison stringComparison = useRegister ? StringComparison.InvariantCulture : StringComparison.InvariantCultureIgnoreCase;

			if (!useRegister)
				text = text.Trim();

			UfeTextLineBase currentLine = _document.GetLine(_endSelection.Row);
			int startColSearch = _endSelection.Col;

			if (_startSelection.Row == _endSelection.Row && Math.Abs(_startSelection.Col - _endSelection.Col) != text.Length)
				startColSearch = Math.Min(_startSelection.Col, _endSelection.Col);

			string currentText = currentLine.TextAsString.Substring(startColSearch);

			if (currentText.IndexOf(text, stringComparison) >= 0)
			{
				int row = _endSelection.Row;

				ShowLine(row);

				int startIndex = currentLine.TextAsString.Length - (currentText.Length - currentText.IndexOf(text, stringComparison));
				int endIndex = startIndex + text.Length;

				SetPosition(new CharPosition(startIndex, row), new CharPosition(endIndex, row));
				return;
			}

			for (int index = _endSelection.Row + 1; index < _document.LineCount; index++)
			{
				UfeTextLineBase line = _document.GetLine(index);

				if (line.TextAsString.IndexOf(text, stringComparison) >= 0)
				{
					ShowLine(index);

					int startIndex = line.TextAsString.IndexOf(text, stringComparison);
					int endIndex = startIndex + text.Length;

					SetPosition(new CharPosition(startIndex, index), new CharPosition(endIndex, index));
					return;
				}
			}

			for (int index = 0; index < _endSelection.Row; index++)
			{
				UfeTextLineBase line = _document.GetLine(index);

				if (line.TextAsString.IndexOf(text, stringComparison) >= 0)
				{
					ShowLine(index);

					int startIndex = line.TextAsString.IndexOf(text, stringComparison);
					int endIndex = startIndex + text.Length;

					SetPosition(new CharPosition(startIndex, index), new CharPosition(endIndex, index));
					return;
				}
			}
		}

		public void ShowLine(int number)
		{
			_renderedStartLineIndex = Math.Min(Math.Max(0, number - _renderedLineCount/2), _document.LineCount - 1);
			SetPosition(new CharPosition(0, Math.Min(Math.Max(0, number), _document.LineCount - 1)));
		}
		
		protected override void OnPreviewKeyDown(KeyEventArgs e)
		{
			OnKeyDownEngine(e);
		}

		public void OnKeyDownEngine(KeyEventArgs e)
		{
			if (OnKeyDownEvent != null)
			{
				OnKeyDownEvent(e);

				if (e.Handled)
					return;
			}

			bool hasCtrl = (e.KeyboardDevice.Modifiers & ModifierKeys.Control) > 0;
			bool hasAlt = (e.KeyboardDevice.Modifiers & ModifierKeys.Alt) > 0;
			bool hasShift = (e.KeyboardDevice.Modifiers & ModifierKeys.Shift) > 0;

			if (e.Key == Key.F && hasCtrl)
			{
				if (_startSelection.Row == _endSelection.Row)
				{
					UfeTextLineBase line = _document.GetLine(_endSelection.Row);
					string text = line.TextAsString.Substring(Math.Min(_startSelection.Col, _endSelection.Col), Math.Abs(_startSelection.Col - _endSelection.Col));

					_findWordCtrl.SetText(text);
				}
				else
				{
					_findWordCtrl.SetText(string.Empty);
				}
				
				SetTopPanelVisibility(TopPanelVisibility.Word);
				Keyboard.Focus(_findWordCtrl.txtWord);

				e.Handled = true;
			}
			else if (e.Key == Key.I && hasCtrl)
			{
				_findWordCtrl.SetText(string.Empty, true);
				
				SetTopPanelVisibility(TopPanelVisibility.Word);
				Keyboard.Focus(_findWordCtrl.txtWord);

				e.Handled = true;
			}
			else if (e.Key == Key.G && hasCtrl)
			{
				_findLineCtrl.SetLine(_endSelection.Row);
				SetTopPanelVisibility(TopPanelVisibility.Line);
				Keyboard.Focus(_findLineCtrl.txtNumber);

				e.Handled = true;
			}
			else if (e.Key == Key.Tab && !hasCtrl && !hasAlt)
			{
				if (_isReadonly)
					return;

				OnRemoveSelection();

				UfeTextLineBase line = _document.GetLine(_startSelection.Row);
				line.Insert(_startSelection.Col, "\t");
				SetPosition(new CharPosition(_startSelection.Col + 1, _startSelection.Row));
				_document.SetIsChangedToTrue();

				if (_longestLine == line)
					_longestLineWidth = _longestLine.GetWidthInPixels();

				e.Handled = true;
			}
			else if (e.Key == Key.D && hasCtrl)
			{
				if (_isReadonly)
					return;

				if (_startSelection == _endSelection)
				{
					_document.InsertLine(_startSelection.Row + 1, _document.GetLine(_startSelection.Row).TextAsString, false);
					SetPosition(new CharPosition(_startSelection.Col, _startSelection.Row + 1));
					_document.SetIsChangedToTrue();
				}
				else
				{
					string selectedText = GetSelectedText();
					string[] selectedLines = selectedText.Split(new[] {Environment.NewLine, "\n"}, StringSplitOptions.None);

					if (selectedLines.Length > 0)
					{
						CharPosition start = _startSelection < _endSelection ? _startSelection : _endSelection;
						CharPosition end = _startSelection > _endSelection ? _startSelection : _endSelection;

						for (int index = 0, col = end.Col; index < selectedLines.Length; index++, col = 0)
						{
							UfeTextLineBase line = _document.GetLine(end.Row + index);
							string pasteText = selectedLines[index];
							line.Insert(col, pasteText);

							if (index != selectedLines.Length - 1)
							{
								string newLine = line.TextAsString.Substring(col + pasteText.Length);
								_document.InsertLine(end.Row + index + 1, newLine, false);

								line.Remove(line.TextAsString.Length - newLine.Length, newLine.Length);
							}
							else
							{
								SetPosition(new CharPosition(end), new CharPosition(col + pasteText.Length, end.Row + index));
							}
						}
					}

//					SetPosition(new CharPosition(_startSelection > _endSelection ? _startSelection : _endSelection), new CharPosition(col, row));
					_document.SetIsChangedToTrue();
				}

				e.Handled = true;
			}
			else if (e.Key == Key.C && hasCtrl)
			{
				OnCopy();
				e.Handled = true;
			}
			else if (e.Key == Key.X && hasCtrl)
			{
				OnCopy();
				OnRemoveSelection();
				_document.SetIsChangedToTrue();

				e.Handled = true;
				InvalidateVisual();
			}
			else if (e.Key == Key.V && hasCtrl)
			{
				if (_isReadonly)
					return;

				string text = Clipboard.GetText();
				string[] lines = text.Split(new[] { Environment.NewLine }, StringSplitOptions.None);

				OnRemoveSelection();

				if (lines.Length > 0)
				{
					for (int index = 0, col = _startSelection.Col; index < lines.Length; index++, col = 0)
					{
						UfeTextLineBase line = _document.GetLine(_startSelection.Row + index);
						string pasteText = lines[index];
						line.Insert(col, pasteText);

						if (index != lines.Length - 1)
						{
							string newLine = line.TextAsString.Substring(col + pasteText.Length);
							_document.InsertLine(_startSelection.Row + index + 1, newLine, false);

							line.Remove(line.TextAsString.Length - newLine.Length, newLine.Length);
						}
						else
						{
							SetPosition(new CharPosition(col + pasteText.Length, _startSelection.Row + index));
						}
					}
				}

				_document.SetIsChangedToTrue();

				_longestLineWidth = -1;
				FindTheLongestLine();

				e.Handled = true;
			}
			else if (e.Key == Key.A && hasCtrl)
			{
				SetPosition(new CharPosition(0, 0), new CharPosition(_document.GetLine(_document.LineCount - 1).Length, _document.LineCount - 1));

				e.Handled = true;
			}
			else if (e.Key == Key.Delete)
			{
				if (_isReadonly)
					return;

				if (_startSelection != _endSelection)
				{
					OnRemoveSelection();
				}
				else
				{
					CharPosition newPosition = GetNextPosition();

					if (_startSelection == newPosition)
					{
						// nothing
					}
					else if (_startSelection.Row != newPosition.Row)
					{
						_document.GetLine(_startSelection.Row).Append(
							_document.GetLine(newPosition.Row).TextAsString);
						_document.RemoveLine(newPosition.Row);
					}
					else
					{
						_document.GetLine(_startSelection.Row).Remove(_startSelection.Col, 1);
					}
				}

				ShowHint();
				_document.SetIsChangedToTrue();

				e.Handled = true;
				InvalidateVisual();
			}
			else if (e.Key == Key.Back)
			{
				if (_isReadonly)
					return;

				if (_startSelection != _endSelection)
				{
					OnRemoveSelection();
				}
				else
				{
					CharPosition newPosition = GetBackPosition();

					if (_startSelection == newPosition)
					{
						// nothing
					}
					else if (_startSelection.Row != newPosition.Row)
					{
						_document.GetLine(newPosition.Row).Append(
							_document.GetLine(_startSelection.Row).TextAsString);
						_document.RemoveLine(_startSelection.Row);

						SetPosition(newPosition);
					}
					else
					{
						_document.GetLine(_startSelection.Row).Remove(_startSelection.Col - 1, 1);
						SetPosition(new CharPosition(_startSelection.Col - 1, _startSelection.Row));
					}
				}

				ShowHint();
				_document.SetIsChangedToTrue();

				e.Handled = true;
			}
			else if (e.Key == Key.Left && hasCtrl && hasShift)
			{
				SetPosition(null, GetBackCtrlPosition());
			
				e.Handled = true;
			}
			else if (e.Key == Key.Left && hasShift)
			{
				SetPosition(null, GetBackPosition());

				e.Handled = true;
			}
			else if (e.Key == Key.Left && hasCtrl)
			{
				if (_startSelection != _endSelection)
				{
					int col = _startSelection < _endSelection ? _startSelection.Col : _endSelection.Col;
					int row = _startSelection < _endSelection ? _startSelection.Row : _endSelection.Row;

					SetPosition(new CharPosition(col, row));
				}
				else
				{
					SetPosition(GetBackCtrlPosition());
				}

				e.Handled = true;
			}
			else if (e.Key == Key.Left)
			{
				if (_startSelection != _endSelection)
				{
					int col = _startSelection < _endSelection ? _startSelection.Col : _endSelection.Col;
					int row = _startSelection < _endSelection ? _startSelection.Row : _endSelection.Row;

					SetPosition(new CharPosition(col, row));
				}
				else
				{
					CharPosition position = GetBackPosition();
					SetPosition(position);
				}

				e.Handled = true;
			}
			else if (e.Key == Key.Right && hasCtrl && hasShift)
			{
				SetPosition(null, GetNextCtrlPosition());

				e.Handled = true;
			}
			else if (e.Key == Key.Right && hasShift)
			{
				SetPosition(null, GetNextPosition());

				e.Handled = true;
			}
			else if (e.Key == Key.Right && hasCtrl)
			{
				if (_startSelection != _endSelection)
				{
					int col = _startSelection > _endSelection ? _startSelection.Col : _endSelection.Col;
					int row = _startSelection > _endSelection ? _startSelection.Row : _endSelection.Row;

					SetPosition(new CharPosition(col, row));
				}
				else
				{
					SetPosition(GetNextCtrlPosition());
				}

				e.Handled = true;
			}
			else if (e.Key == Key.Right)
			{
				if (_startSelection != _endSelection)
				{
					int col = _startSelection > _endSelection ? _startSelection.Col : _endSelection.Col;
					int row = _startSelection > _endSelection ? _startSelection.Row : _endSelection.Row;

					SetPosition(new CharPosition(col, row));
				}
				else
				{
					CharPosition position = GetNextPosition();
					SetPosition(position);
				}

				e.Handled = true;
			}
			else if (e.Key == Key.Up && hasShift)
			{
				SetPosition(null, GetUpPosition());

				e.Handled = true;
			}
			else if (e.Key == Key.Up)
			{
				if (_popup.IsOpen)
				{
					_popup.OnUp();
				}
				else
				{
					SetPosition(GetUpPosition());
					e.Handled = true;
				}
			}
			else if (e.Key == Key.Down && hasShift)
			{
				SetPosition(null, GetDownPosition());

				e.Handled = true;
			}
			else if (e.Key == Key.Down)
			{
				if (_popup.IsOpen)
				{
					_popup.OnDown();
				}
				else
				{
					SetPosition(GetDownPosition());

					e.Handled = true;
				}
			}
			else if (e.Key == Key.Return)
			{
				if (_isReadonly)
					return;

				if (_popup.IsOpen)
				{
					_popup.OnEnter();
					e.Handled = true;
				}
				else
				{
					OnRemoveSelection();

					UfeTextLineBase line = _document.GetLine(_startSelection.Row);
					string text = line.TextAsString.Substring(_startSelection.Col);

					line.Remove(_startSelection.Col, text.Length);

					int row = _startSelection.Row + 1;
					int col = _document.InsertLine(row, text, true);
					SetPosition(new CharPosition(col, row));

					e.Handled = true;
				}

				_document.SetIsChangedToTrue();
			}
			else if (e.Key == Key.Escape)
			{
				SetPosition(_endSelection);

				e.Handled = true;
				_popup.IsOpen = false;
			}
			else if (e.Key == Key.Home && hasShift)
			{
				SetPosition(null, new CharPosition(0, _endSelection.Row));

				e.Handled = true;
			}
			else if (e.Key == Key.Home)
			{
				SetPosition(new CharPosition(0, _endSelection.Row));

				e.Handled = true;
			}
			else if (e.Key == Key.End && hasShift)
			{
				SetPosition(null, new CharPosition(_document.GetLine(_endSelection.Row).Length, _endSelection.Row));

				e.Handled = true;
			}
			else if (e.Key == Key.End)
			{
				SetPosition(new CharPosition(_document.GetLine(_endSelection.Row).Length, _endSelection.Row));
				e.Handled = true;
			}
			else if (e.Key == Key.Oem2 && hasCtrl)
			{
				if (!IsReadonly)
				{
					if (_startSelection == _endSelection)
					{
						UfeTextLineBase line = _document.GetLine(_endSelection.Row);
						
						if (line.CommentLine())
						{
							_document.SetIsChangedToTrue();

							if (_longestLine == line)
								_longestLineWidth = _longestLine.GetWidthInPixels();

							InvalidateVisual();
						}
					}
				}

				e.Handled = true;
			}
			else if (e.Key == Key.Space && hasCtrl)
			{
				if (!IsReadonly)
					ShowHint();

				e.Handled = true;
			}
		}

		private void ShowHint()
		{
			_popup.Commands = _document.GetLine(_endSelection.Row).GetHints(_endSelection.Col);
			_popup.IsOpen = _popup.Commands.Length > 0;
		}

		private void HideHint()
		{
			_popup.IsOpen = false;
		}

		private void SetPosition(CharPosition position)
		{
			SetPosition(position, position);
		}

		private void SetPosition(CharPosition start, CharPosition end)
		{
			bool isChanged = false;

			if (start != null)
			{
				_startSelection = start;
				isChanged = true;
			}

			if (end != null)
			{
				_endSelection = end;
				isChanged = true;
			}

			if (isChanged)
			{
				InvalidateVisual();
				HideHint();
			}
		}

		private CharPosition GetBackCtrlPosition()
		{
			int position = _endSelection.Col;

			if (position == 0)
				return GetBackPosition();

			UfeTextLineBase line = _document.GetLine(_endSelection.Row);

			string text = line.TextAsString.Substring(0, position).TrimEnd(CtrlSplitChars);
			int newIndex = Math.Max(text.LastIndexOfAny(CtrlSplitChars) + 1, 0);
			return new CharPosition(newIndex, _endSelection.Row);
		}

		private CharPosition GetNextCtrlPosition()
		{
			UfeTextLineBase line = _document.GetLine(_endSelection.Row);
			int position = _endSelection.Col;

			if (position == line.Length)
				return GetNextPosition();

			string text = line.TextAsString.Substring(position, line.TextAsString.Length - position).TrimStart(CtrlSplitChars);
			int firstIndex = text.IndexOfAny(CtrlSplitChars);

			if (firstIndex < 0)
				return new CharPosition(line.TextAsString.Length, _endSelection.Row);

			int newIndex = line.TextAsString.Length - (text.Length - firstIndex);
			newIndex = line.TextAsString.Length - line.TextAsString.Substring(newIndex).TrimStart(CtrlSplitChars).Length;
			return new CharPosition(newIndex, _endSelection.Row);
		}

		private CharPosition GetDownPosition()
		{
			int col = _endSelection.Col;
			int row = _endSelection.Row;

			if (row < _document.LineCount - 1)
				row++;

			col = Math.Min(col, _document.GetLine(row).Length);

			return new CharPosition(col, row);
		}

		private CharPosition GetUpPosition()
		{
			int col = _endSelection.Col;
			int row = _endSelection.Row;

			if (row > 0)
				row--;

			col = Math.Min(col, _document.GetLine(row).Length);

			return new CharPosition(col, row);
		}

		private CharPosition GetBackPosition()
		{
			int col = _endSelection.Col;
			int row = _endSelection.Row;

			if (col > 0)
				col--;
			else if (row > 0)
			{
				row--;
				col = _document.GetLine(row).Length;
			}

			return new CharPosition(col, row);
		}

		private CharPosition GetNextPosition()
		{
			int col = _endSelection.Col;
			int row = _endSelection.Row;

			if (col < _document.GetLine(row).Length)
				col++;
			else if (row < _document.LineCount - 1)
			{
				row++;
				col = 0;
			}

			return new CharPosition(col, row);
		}

		private void OnRemoveSelection()
		{
			_document.Remove(_startSelection, _endSelection);

			int newCol = _startSelection < _endSelection ? _startSelection.Col : _endSelection.Col;
			int newRow = _startSelection < _endSelection ? _startSelection.Row : _endSelection.Row;

			SetPosition(new CharPosition(newCol, newRow));
			_document.SetIsChangedToTrue();
		}

		private void OnCopy()
		{
			DataObject data = new DataObject();

			CopyAsText(data);
			CopyAsRtf(data);

			Clipboard.SetDataObject(data, true);
		}

		private void CopyAsRtf(DataObject data)
		{
			FlowDocument flowDocument = new FlowDocument();

			if (_startSelection.Row == _endSelection.Row)
			{
				Paragraph line = new Paragraph();
				flowDocument.Blocks.Add(line);

				foreach (UfeTextStyle style in _document.GetLine(_startSelection.Row).Substring(Math.Min(_startSelection.Col, _endSelection.Col), Math.Abs(_startSelection.Col - _endSelection.Col)))
				{
					line.Inlines.Add(new Run(style.Text.ToString()) { Foreground = style.Brush, FontFamily = Typeface.FontFamily, FontSize = fontSize, FontStyle = Typeface.Style, FontWeight = Typeface.Weight, FontStretch = Typeface.Stretch });
				}
			}
			else
			{
				CharPosition start = _startSelection < _endSelection ? _startSelection : _endSelection;
				CharPosition end = start == _startSelection ? _endSelection : _startSelection;

				for (int index = start.Row; index <= end.Row; index++)
				{
					UfeTextLineBase line = _document.GetLine(index);

					Paragraph paragraph = new Paragraph();
					flowDocument.Blocks.Add(paragraph);

					if (index == start.Row)
					{
						foreach (UfeTextStyle style in line.Substring(start.Col))
						{
							paragraph.Inlines.Add(new Run(style.Text.ToString()) { Foreground = style.Brush, FontFamily = Typeface.FontFamily, FontSize = fontSize, FontStyle = Typeface.Style, FontWeight = Typeface.Weight, FontStretch = Typeface.Stretch });
						}
					}
					else if (index == end.Row)
					{
						foreach (UfeTextStyle style in line.Substring(0, end.Col))
						{
							paragraph.Inlines.Add(new Run(style.Text.ToString()) { Foreground = style.Brush, FontFamily = Typeface.FontFamily, FontSize = fontSize, FontStyle = Typeface.Style, FontWeight = Typeface.Weight, FontStretch = Typeface.Stretch });
						}
					}
					else
					{
						foreach (UfeTextStyle style in line.Text)
						{
							paragraph.Inlines.Add(new Run(style.Text.ToString()) { Foreground = style.Brush, FontFamily = Typeface.FontFamily, FontSize = fontSize, FontStyle = Typeface.Style, FontWeight = Typeface.Weight, FontStretch = Typeface.Stretch });
						}
					}
				}
			}

			TextRange content = new TextRange(flowDocument.ContentStart, flowDocument.ContentEnd);
			StringBuilder result = new StringBuilder();

			using (MemoryStream stream = new MemoryStream())
			{
				content.Save(stream, DataFormats.Rtf);
				stream.Flush();
				stream.Position = 0;
				result.Append(new StreamReader(stream).ReadToEnd());
			}

			data.SetText(result.ToString(), TextDataFormat.Rtf);
		}

		private string GetSelectedText()
		{
			StringBuilder builder = new StringBuilder();

			CharPosition start = _startSelection < _endSelection ? _startSelection : _endSelection;
			CharPosition end = start == _startSelection ? _endSelection : _startSelection;

			if (start == end)
			{
				return string.Empty;
			}
			else if (start.Row == end.Row)
			{
				return _document.GetLine(start.Row).TextAsString.Substring(start.Col, end.Col - start.Col);
			}
			else
			{
				for (int index = start.Row; index <= end.Row; index++)
				{
					UfeTextLineBase line = _document.GetLine(index);

					if (index == start.Row)
					{
						builder.AppendLine(line.TextAsString.Substring(start.Col));
					}
					else if (index == end.Row)
					{
						builder.Append(line.TextAsString.Substring(0, end.Col));
					}
					else
					{
						builder.AppendLine(line.TextAsString);
					}
				}
			}

			return builder.ToString();
		}

		private void CopyAsText(DataObject data)
		{
			string selectedString = GetSelectedText();

			data.SetText(selectedString, TextDataFormat.Text);
			data.SetText(selectedString, TextDataFormat.UnicodeText);
		}

		private Tuple<CharPosition, CharPosition> GetWordSelection(CharPosition middleOfWord)
		{
			UfeTextLineBase line = _document.GetLine(middleOfWord.Row);
			int start = Math.Max(line.TextAsString.LastIndexOfAny(CtrlSplitChars, Math.Min(middleOfWord.Col, line.TextAsString.Length - 1)) + 1, 0);
			int end = line.TextAsString.IndexOfAny(CtrlSplitChars, middleOfWord.Col);

			if (end < 0)
				end = line.TextAsString.Length;

			return new Tuple<CharPosition, CharPosition>(new CharPosition(start, middleOfWord.Row), new CharPosition(end, middleOfWord.Row));
		}

		protected override void OnMouseMove(MouseEventArgs e)
		{
			if (e.LeftButton == MouseButtonState.Pressed)
			{
				Point location = e.GetPosition(this);
				CharPosition temp = GetChartPosition(location);

				if (temp != null)
				{
					if (_isLastMouseDownDouble)
					{
						Tuple<CharPosition, CharPosition> startSelection = GetWordSelection(_lastMouseDownPosition);
						Tuple<CharPosition, CharPosition> endSelection = GetWordSelection(temp);

						if (temp < _lastMouseDownPosition)
						{
							SetPosition(startSelection.Item2, endSelection.Item1);
						}
						else
						{
							SetPosition(startSelection.Item1, endSelection.Item2);

						}
					}
					else
					{
						SetPosition(null, temp);
					}
				}
			}

			e.Handled = true;
		}
		
		protected override void OnMouseLeftButtonUp(MouseButtonEventArgs e)
		{
			if (_isLastMouseDownDouble)
			{
				/*Point location = e.GetPosition(this);
				CharPosition temp = GetChartPosition(location);

				if (temp != null)
				{
					UfeTextLineBase line = _document.GetLine(temp.Row);
//					int start = Math.Max(line.TextAsString.LastIndexOfAny(CtrlSplitChars, Math.Min(temp.Col, line.TextAsString.Length - 1)) + 1, 0);
					int end = line.TextAsString.IndexOfAny(CtrlSplitChars, temp.Col);

					if (end < 0)
						end = line.TextAsString.Length;

					SetPosition(null, new CharPosition(end, temp.Row));
				}*/
			}
			else
			{
				Point location = e.GetPosition(this);
				CharPosition temp = GetChartPosition(location);

				if (temp != null)
					SetPosition(null, temp);
			}

			e.Handled = true;
			Keyboard.Focus(this);
			ReleaseMouseCapture();
		}

		private class TextPosition
		{
			private readonly string _textAsString;
			private readonly FormattedText _text;
			private readonly double _top;
			private readonly double _buttom;
			private readonly int _line;

			public TextPosition(string textAsString, FormattedText text, double top, double buttom, int line)
			{
				if (text == null) 
					throw new ArgumentNullException("text");

				_textAsString = textAsString;
				_text = text;
				_top = top;
				_buttom = buttom;
				_line = line;
			}

			public string TextAsString
			{
				get { return _textAsString; }
			}

			public int Line
			{
				get { return _line; }
			}

			public FormattedText Text
			{
				get { return _text; }
			}

			public double Top
			{
				get { return _top; }
			}

			public double Buttom
			{
				get { return _buttom; }
			}
		}
	}
}
