﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Globalization;
using System.Linq;
using System.Text;
using System.Threading;
using System.Windows;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using TextEditor.Properties;
using UniversalEditor.Base.FileSystem;
using UniversalEditor.Base.Options;
using UniversalEditor.Base.Utils.LoggerModule;
using Brush = System.Windows.Media.Brush;
using SystemColors = System.Windows.SystemColors;

namespace TextEditor
{
	public class UfeTextDocument
	{
		public event ThreadStart ContentChanged;
		public event ThreadStart FileDeleted;

		private readonly FileWrapper _file;
		private readonly object _arg;
		private readonly List<UfeTextLineBase> _lines = new List<UfeTextLineBase>();
		
		public void SetIsChangedToTrue()
		{
			_file.SetIsChangedToTrue();
		}

		public UfeTextDocument(FileWrapper file, object arg)
		{
			if (file == null) 
				throw new ArgumentNullException("file");

			_file = file;
			_arg = arg;

			ReloadContentInner();

			_file.ContentChanged += OnContentChanged;
			_file.FileDeleted += OnFileDeleted;
		}

		private void ReloadContentInner()
		{
			_lines.Clear();

			// add lines
			string[] lines = _file.AllText.ToString().Split(new[] {"\r\n", "\n", Environment.NewLine}, StringSplitOptions.None);

			for (int index = 0; index < lines.Length; index++)
				_lines.Add(ParseLine(lines[index], _arg));

			// set prev & next lines
			if (_lines.Count > 1)
			{
				_lines[0].SetNextLine(_lines[1]);
			}

			for (int index = 1; index < _lines.Count - 1; index++)
			{
				_lines[index].SetPrevLine(_lines[index - 1]);
				_lines[index].SetNextLine(_lines[index + 1]);
			}

			if (_lines.Count > 2)
			{
				_lines[_lines.Count - 1].SetPrevLine(_lines[_lines.Count - 2]);
			}

			// initialize
			for (int index = 0; index < _lines.Count; index++)
				_lines[index].Initialize();
		}

		private void OnFileDeleted()
		{
			if (FileDeleted != null)
				FileDeleted();
		}

		private void OnContentChanged()
		{
			if (ContentChanged != null)
				ContentChanged();
		}

		public bool IsNewFile
		{
			get { return string.IsNullOrEmpty(_file.FilePath); }
		}

		public int LineCount
		{
			get { return _lines.Count; }
		}

		public UfeTextLineBase GetLine(int index)
		{
			return _lines[index];
		}

		public int InsertLine(int index, string text, bool addTabsToStart)
		{
			StringBuilder prefix = new StringBuilder();

			if (addTabsToStart && index > 0)
			{
				string prevText = GetLine(index - 1).TextAsString;

				for (int marker = 0; marker < prevText.Length; marker++)
				{
					char ch = prevText[marker];

					if (ch == ' ' || ch == '\t')
						prefix.Append(ch);
					else break;
				}
			}

			UfeTextLineBase line = ParseLine(prefix + text, _arg);

			UfeTextLineBase prevLine = null;
			UfeTextLineBase nextLine = null;

			if (index > 0)
			{
				prevLine = _lines[index - 1];
				line.SetPrevLine(prevLine);
				prevLine.SetNextLine(line);
			}

			if (index < _lines.Count)
			{
				nextLine = _lines[index];
				line.SetNextLine(nextLine);
				nextLine.SetPrevLine(line);
			}

			_lines.Insert(index, line);

			if (prevLine != null)
				prevLine.Initialize();
			
			line.Initialize();
			
			if (nextLine != null)
				nextLine.Initialize();

			return prefix.Length;
		}

		public void Save(string filePath)
		{
			_file.AllText.Clear();

			for (int index = 0; index < _lines.Count - 1; index++)
				_file.AllText.AppendLine(_lines[index].TextAsString);

			_file.AllText.Append(_lines.Last().TextAsString);

			_file.SaveAs(filePath);
		}

		public void SaveCopy(string filePath)
		{
			_file.AllText.Clear();

			for (int index = 0; index < _lines.Count - 1; index++)
				_file.AllText.AppendLine(_lines[index].TextAsString);

			_file.AllText.Append(_lines.Last().TextAsString);

			_file.SaveCopyAs(filePath);
		}

		protected virtual UfeTextLineBase ParseLine(string text, object arg)
		{
			return new UfeTextLine(text);
		}

		public void Remove(CharPosition startSelection, CharPosition endSelection)
		{
			CharPosition begin = startSelection < endSelection ? startSelection : endSelection;
			CharPosition end = begin == startSelection ? endSelection : startSelection;

			if (begin.Row == end.Row)
			{
				UfeTextLineBase line = _lines[begin.Row];
				line.Remove(begin.Col, end.Col - begin.Col);
			}
			else
			{
				for (int index = end.Row - 1; index > begin.Row; index--)
				{
					_lines.RemoveAt(index);
				}

				UfeTextStyle[] lastPart = _lines[begin.Row + 1].Substring(end.Col);
				_lines.RemoveAt(begin.Row + 1);

				UfeTextLineBase line = _lines[begin.Row];
				line.Remove(begin.Col, line.Length - begin.Col);
				line.Append(lastPart);
			}
		}

		public void RemoveLine(int row)
		{
			_lines.RemoveAt(row);
		}

		public void Reload()
		{
			_file.Reload();
			ReloadContentInner();
		}
	}

	public abstract class UfeTextLineBase
	{
		private UfeTextLineBase _prevLine;
		private UfeTextLineBase _nextLine;

		public abstract void Remove(int start, int lenght);
		public abstract UfeTextStyle[] Substring(int start, int lenght);
		public abstract UfeTextStyle[] Substring(int start);
		public abstract UfeTextStyle[] Text { get; }

		public abstract int Length { get; }
		public abstract void Append(string text);
		public abstract void Append(UfeTextStyle[] parts);
		public abstract void Insert(int start, string text);

		public abstract HintItem[] GetHints(int position);

		public abstract string TextAsString { get; }

		internal double GetWidthInPixels()
		{
			FormattedText formattedText = new FormattedText(UfeTextBoxInner.ReplaceTabs(TextAsString), CultureInfo.CurrentUICulture, FlowDirection.LeftToRight, UfeTextBoxInner.Typeface,
				Options.Instance.FontSize, Application.Current.MainWindow.Foreground) { Trimming = TextTrimming.None };

			return formattedText.Width;
		}

		internal void SetPrevLine(UfeTextLineBase line)
		{
			_prevLine = line;
		}

		internal void SetNextLine(UfeTextLineBase line)
		{
			_nextLine = line;
		}

		protected UfeTextLineBase PrevLine
		{
			get { return _prevLine; }
		}

		protected UfeTextLineBase NextLine
		{
			get { return _nextLine; }
		}

		public abstract void Initialize();

		public virtual bool CommentLine()
		{
			return false;
		}

		public virtual bool IsRendarableInReadonly()
		{
			return true;
		}
	}

	public class HintItem
	{
		public enum ImageType { Default, Method, Variable }

		private readonly string _value;
		private readonly string _prefix;
		private readonly string _displayName;
		private readonly ImageType _imageType;

		public HintItem(string value)
			: this(value, ImageType.Default, value, string.Empty)
		{}

		public HintItem(string displayName, ImageType imageType, string value, string prefix)
		{
			_value = value;
			_prefix = prefix;
			_displayName = displayName;
			_imageType = imageType;
		}

		public string Value
		{
			get { return _value; }
		}

		public string Prefix
		{
			get { return _prefix; }
		}

		public string DisplayName
		{
			get { return _displayName; }
		}

		public ImageSource Image
		{
			get
			{
				if (_imageType == ImageType.Method)
					return Resources.hint_method.ToImageSource();

				if (_imageType == ImageType.Variable)
					return Resources.hint_variable.ToImageSource();

				return Resources.hint_default.ToImageSource();
			}
		}

		public override string ToString()
		{
			return DisplayName;
		}
	}

	public class UfeTextStyle
	{
		private readonly Brush _brush;
		private readonly StringBuilder _text;

		public UfeTextStyle(string text, Brush brush)
		{
			_brush = brush;
			_text = new StringBuilder(text);
		}

		public Brush Brush
		{
			get { return _brush; }
		}

		public StringBuilder Text
		{
			get { return _text; }
		}

		public override string ToString()
		{
			return _text.ToString();
		}
	}

	public class UfeTextLine : UfeTextLineBase
	{
		private readonly UfeTextStyle _text;

		public UfeTextLine(string text)
		{
			_text = new UfeTextStyle(text, SystemColors.ControlTextBrush);
		}

		public override void Initialize()
		{}

		public override void Remove(int start, int lenght)
		{
			_text.Text.Remove(start, lenght);
		}

		public override UfeTextStyle[] Substring(int start, int lenght)
		{
			return new[] { new UfeTextStyle(_text.Text.ToString().Substring(start, lenght), SystemColors.ControlTextBrush) };
		}

		public override UfeTextStyle[] Substring(int start)
		{
			return new[] { new UfeTextStyle(_text.Text.ToString().Substring(start), SystemColors.ControlTextBrush) };
		}

		public override UfeTextStyle[] Text
		{
			get { return new[] {_text}; }
		}

		public override int Length
		{
			get { return _text.Text.Length; }
		}

		public override void Append(string text)
		{
			_text.Text.Append(text);
		}

		public override void Append(UfeTextStyle[] parts)
		{
			foreach (UfeTextStyle ufeTextStyle in parts)
			{
				Append(ufeTextStyle.Text.ToString());
			}
		}

		public override void Insert(int start, string text)
		{
			_text.Text.Insert(start, text);
		}

		public override HintItem[] GetHints(int position)
		{
			return new HintItem[0];
		}

		public override string TextAsString
		{
			get { return _text.Text.ToString(); }
		}

		public override string ToString()
		{
			return _text.Text.ToString();
		}
	}

	public class CharPosition
	{
		private readonly int _col;
		private readonly int _row;

		public CharPosition(CharPosition source)
			: this(source.Col, source.Row)
		{}

		public CharPosition(int col, int row)
		{
			_col = col;
			_row = row;
		}

		public int Col
		{
			get { return _col; }
		}

		public int Row
		{
			get { return _row; }
		}

		public static bool operator <(CharPosition first, CharPosition second)
		{
			return first.Row < second.Row || (first.Row == second.Row && first.Col < second.Col);
		}

		public static bool operator >(CharPosition first, CharPosition second)
		{
			return first.Row > second.Row || (first.Row == second.Row && first.Col > second.Col);
		}
		public static bool operator ==(CharPosition first, CharPosition second)
		{
			if (ReferenceEquals(first, null) && ReferenceEquals(second, null))
				return true;
			if (ReferenceEquals(first, null) || ReferenceEquals(second, null))
				return false;

			return first.Row == second.Row && first.Col == second.Col;
		}

		public static bool operator !=(CharPosition first, CharPosition second)
		{
			return !(first == second);
		}

		public override string ToString()
		{
			return string.Format("Row {0}, Col {1}", _row, _col);
		}
	}
}
