﻿using System;
using System.IO;
using System.Text;
using System.Threading;
using System.Windows;

namespace UniversalEditor.Base.FileSystem
{
	public class FileWrapper
	{
		public event ThreadStart ContentChanged;
		public event ThreadStart FileDeleted;
		public event ThreadStart HasChangesChanged;

		private readonly FileSystemWatcher _fileWatcher = new FileSystemWatcher();
		private readonly Timer _timer;

		private string _filePath = string.Empty;
		private bool _isNew;
		private bool _hasChanges = false;
		private StringBuilder _fileContent = null;

		public bool HasChanges
		{
			get { return _hasChanges; }
			private set
			{
				if (_hasChanges == value)
					return;

				_hasChanges = value;

				if (HasChangesChanged != null)
					HasChangesChanged();
			}
		}

		public void SetIsChangedToTrue()
		{
			HasChanges = true;
		}

		public string FilePath
		{
			get { return _isNew ? string.Empty : _filePath; }
		}

		public FileWrapper(string filePath, bool isNew)
		{
			_timer = new Timer(OnFileChangedRaised);

			_filePath = filePath;
			_isNew = isNew;
			_hasChanges = isNew;

			if (!string.IsNullOrEmpty(filePath) && !isNew)
				InitializeWatcher();
		}

		private void OnFileChangedRaised(object state)
		{
			if (ContentChanged != null)
				ContentChanged();
		}

		private void OnFileDeleted(object sender, FileSystemEventArgs e)
		{
			if (FileDeleted != null)
				FileDeleted();
		}

		private void OnFileRenamed(object sender, RenamedEventArgs e)
		{
			if (e.OldName == Path.GetFileName(_filePath))
			{
				if (FileDeleted != null)
					FileDeleted();
			}
			else
			{
				if (ContentChanged != null)
					ContentChanged();
			}
		}

		private void OnFileChanged(object sender, FileSystemEventArgs e)
		{
			_timer.Change(500, int.MaxValue);
		}

		public StringBuilder AllText
		{
			get
			{
				if (_fileContent == null)
				{
					_fileContent = string.IsNullOrEmpty(_filePath) ? new StringBuilder() : new StringBuilder(File.ReadAllText(_filePath));
				}

				return _fileContent;
			}
		}

		public void Save()
		{
			if (_fileContent == null)
				return;

			_fileWatcher.EnableRaisingEvents = false;
			File.WriteAllText(_filePath, _fileContent.ToString());
			_fileWatcher.EnableRaisingEvents = true;

			_isNew = false;
			HasChanges = false;
		}

		public void Dispose()
		{
			_fileWatcher.Dispose();
		}

		private void InitializeWatcher()
		{
			_fileWatcher.Path = Path.GetDirectoryName(_filePath);
			_fileWatcher.Filter = Path.GetFileName(_filePath);
			_fileWatcher.Changed += OnFileChanged;
			_fileWatcher.Created += OnFileChanged;
			_fileWatcher.Deleted += OnFileDeleted;
			_fileWatcher.Renamed += OnFileRenamed;
			_fileWatcher.EnableRaisingEvents = true;
		}

		private void UnitializeWatcher()
		{
			_fileWatcher.EnableRaisingEvents = false;
			_fileWatcher.Changed -= OnFileChanged;
			_fileWatcher.Created -= OnFileChanged;
			_fileWatcher.Deleted -= OnFileDeleted;
			_fileWatcher.Renamed -= OnFileRenamed;
		}

		public void SaveAs(string filePath)
		{
			if (string.IsNullOrEmpty(filePath))
				throw new ArgumentNullException("filePath");

			if (!string.IsNullOrEmpty(_filePath) && !_isNew)
				UnitializeWatcher();

			_filePath = filePath;

			string dir = Path.GetDirectoryName(_filePath);

			if (!Directory.Exists(dir))
				Directory.CreateDirectory(dir);

			File.WriteAllText(_filePath, AllText.ToString());
			
			InitializeWatcher();

			_isNew = false;
			HasChanges = false;
		}

		public void SaveCopyAs(string filePath)
		{
			if (string.IsNullOrEmpty(filePath))
				throw new ArgumentNullException("filePath");

			string dir = Path.GetDirectoryName(filePath);

			if (!Directory.Exists(dir))
				Directory.CreateDirectory(dir);

			File.WriteAllText(filePath, AllText.ToString());
		}

		public void Reload()
		{
			_fileContent = null;
			HasChanges = false;
		}
	}
}
