﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using System.Windows;
using System.Windows.Media;
using TextEditor;
using UniversalEditor.Base.FileSystem;

namespace UniversalEditor.Xml
{
	internal class XmlTextDocument : UfeTextDocument
	{
		public XmlTextDocument(FileWrapper file)
			: base(file, null)
		{ }

		protected override UfeTextLineBase ParseLine(string text, object arg)
		{
			return new XmlTextLine(text);
		}

		private class XmlTextLine : UfeTextLineBase
		{
			private readonly StringBuilder _text;
			private readonly List<UfeTextStyle> _parts = new List<UfeTextStyle>();

			private bool _endsWithOpenedTag;
			private bool _endsWithOpenedHeader;
			private bool _endsWithComment;

			private bool _isRenderableUnderReadonly = true;

			public XmlTextLine(string text)
			{
				_text = new StringBuilder(text);
			}

			public override bool IsRendarableInReadonly()
			{
				return _isRenderableUnderReadonly;
			}

			public override void Initialize()
			{
				_parts.Clear();

				_isRenderableUnderReadonly = !string.IsNullOrWhiteSpace(_text.ToString());

				if (_text.Length == 0)
				{
					_parts.Add(new UfeTextStyle(string.Empty, SystemColors.ControlTextBrush));
					_endsWithOpenedTag = PrevLine != null && ((XmlTextLine) PrevLine)._endsWithOpenedTag;
					_endsWithOpenedHeader = PrevLine != null && ((XmlTextLine)PrevLine)._endsWithOpenedHeader;
					_endsWithComment = PrevLine != null && ((XmlTextLine)PrevLine)._endsWithComment;
					return;
				}

				List<Tuple<char, Brush>> word = new List<Tuple<char, Brush>>();
				bool isTag = PrevLine != null && ((XmlTextLine)PrevLine)._endsWithOpenedTag;
				bool isHeaderTag = PrevLine != null && ((XmlTextLine)PrevLine)._endsWithOpenedHeader;
				bool isComment = PrevLine != null && ((XmlTextLine)PrevLine)._endsWithComment;
				bool isString = false;

				for (int index = 0; index < _text.Length; index++)
				{
					char ch = _text[index];

					if (ch == '<' && IsChar(_text, index+1, '!') && IsChar(_text, index+2, '-') && IsChar(_text, index+3, '-'))
					{
						word.Add(new Tuple<char, Brush>('<', Brushes.Green));
						word.Add(new Tuple<char, Brush>('!', Brushes.Green));
						word.Add(new Tuple<char, Brush>('-', Brushes.Green));
						word.Add(new Tuple<char, Brush>('-', Brushes.Green));

						isComment = true;
						index += 3;
					}
					else if (ch == '-' && IsChar(_text, index+1, '-') && IsChar(_text, index+2, '>'))
					{
						word.Add(new Tuple<char, Brush>('-', Brushes.Green));
						word.Add(new Tuple<char, Brush>('-', Brushes.Green));
						word.Add(new Tuple<char, Brush>('>', Brushes.Green));

						isComment = false;
						index += 2;
					}
					else if (isComment)
					{
						word.Add(new Tuple<char, Brush>(ch, Brushes.Green));
					}
					else if (ch == '<' && IsChar(_text, index + 1, '?'))
					{
						word.Add(new Tuple<char, Brush>('<', Brushes.Blue));
						word.Add(new Tuple<char, Brush>('?', Brushes.Blue));
						
						isHeaderTag = true;
						index++;
					}
					else if (ch == '<')
					{
						word.Add(new Tuple<char, Brush>(ch, Brushes.Blue));
						isTag = true;
					}
					else if (ch == '?' && IsChar(_text, index + 1, '>'))
					{
						word.Add(new Tuple<char, Brush>('?', Brushes.Blue));
						word.Add(new Tuple<char, Brush>('>', Brushes.Blue));

						isHeaderTag = false;
						index++;
					}
					else if (ch == '>')
					{
						word.Add(new Tuple<char, Brush>(ch, Brushes.Blue));
						isTag = false;
					}
					else if (ch == '"')
					{
						word.Add(new Tuple<char, Brush>(ch, Brushes.Blue));
						isString = !isString;
					}
					else if (ch == '/' || ch == '=')
					{
						word.Add(new Tuple<char, Brush>(ch, Brushes.Blue));
					}
					else if (isHeaderTag)
					{
						word.Add(new Tuple<char, Brush>(ch, Brushes.Blue));
					}
					else
					{
						if (isTag && !isString)
							word.Add(new Tuple<char, Brush>(ch, Brushes.Brown));
						else
							word.Add(new Tuple<char, Brush>(ch, Brushes.Black));
					}
				}

				bool isTagChanged = _endsWithOpenedTag != isTag || _endsWithOpenedHeader != isHeaderTag || _endsWithComment != isComment;

				_endsWithOpenedTag = isTag;
				_endsWithOpenedHeader = isHeaderTag;
				_endsWithComment = isComment;

				_parts.Add(new UfeTextStyle(string.Empty, word.First().Item2));

				for (int index = 0; index < word.Count; index++)
				{
					Tuple<char, Brush> ch = word[index];
					UfeTextStyle lastPart = _parts.Last();

					if (lastPart.Brush.Equals(ch.Item2))
						lastPart.Text.Append(ch.Item1);
					else
					{
						_parts.Add(new UfeTextStyle(ch.Item1.ToString(), ch.Item2));
					}
				}

				if (isTagChanged && NextLine != null)
					NextLine.Initialize();
			}

			public override bool CommentLine()
			{
				if (_text.ToString().Trim().StartsWith("<!--") && _text.ToString().Trim().EndsWith("-->"))
				{
					int startIndex = _text.ToString().IndexOf("<!--");
					_text.Remove(startIndex, 4);

					int endIndex = _text.ToString().IndexOf("-->");
					_text.Remove(endIndex, 3);
				}
				else
				{
					int startIndex = _text.Length - _text.ToString().TrimStart(' ', '\t').Length;
					_text.Insert(startIndex, "<!--");
					_text.Append("-->");
				}
				
				Initialize();
				return true;
			}

			private static bool IsChar(StringBuilder text, int index, char ch)
			{
				if (index >= text.Length)
					return false;

				return text[index] == ch;
			}

			public override void Remove(int start, int lenght)
			{
				_text.Remove(start, lenght);
				Initialize();
			}

			public override UfeTextStyle[] Substring(int start, int lenght)
			{
				List<UfeTextStyle> result = new List<UfeTextStyle>();

				for (int index = 0; index < _parts.Count; index++)
				{
					UfeTextStyle part = _parts[index];

					if (part.Text.Length < start)
					{
						start -= part.Text.Length;
						continue;
					}

					if (part.Text.Length > start + lenght)
					{
						result.Add(new UfeTextStyle(part.Text.ToString().Substring(start, lenght), part.Brush));
						break;
					}

					int l = part.Text.Length - start;
					result.Add(new UfeTextStyle(part.Text.ToString().Substring(start, l), part.Brush));
					lenght -= l;
					start = 0;
				}

				return result.ToArray();
			}

			public override UfeTextStyle[] Substring(int start)
			{
				return Substring(start, Length - start);
			}

			public override UfeTextStyle[] Text
			{
				get { return _parts.ToArray(); }
			}

			public override int Length
			{
				get { return _text.Length; }
			}

			public override void Append(string text)
			{
				_text.Append(text);
				Initialize();
			}

			public override void Append(UfeTextStyle[] parts)
			{
				Array.ForEach(parts, x => _text.Append(x.Text));
				Initialize();
			}

			public override void Insert(int start, string text)
			{
				_text.Insert(start, text);
				Initialize();
			}

			public override HintItem[] GetHints(int position)
			{
				return new HintItem[0];
			}

			public override string TextAsString
			{
				get { return _text.ToString(); }
			}
		}
	}
}
