﻿using System;
using System.Collections.ObjectModel;
using System.IO;
using System.Linq;
using System.Threading;
using UniversalEditor.Base;
using UniversalEditor.Base.FileSystem;
using UniversalEditor.Formats;
using UniversalEditor.PlainText;

namespace UniversalEditor
{
	internal class DomainModel : IEditorOwner
	{
		internal event ThreadStart SelectedEditorChanged;
		internal event ThreadStart HeaderChanged;
		internal event ThreadStart AllHeaderChanged;

		private readonly IFormat[] _editors = new IFormat[]
		{
			new PlainTextFormat(), 
			new PlainImageFormat(),
			new CsvFormat(), 
			new PdfFormat(),
 			new OpenXmlFormat(), 
			new VideoFormat(), 
			new IconFormat(), 
			new AudioFormat(), 
			new XmlFormat(),
			new BatFormat(), 
			new NsisFormat(),
			new CsharpFormat(), 
			new JavaFormat(),
			new ActionscriptFormat(), 
		};

		private readonly ObservableCollection<EditorBase> _openedEditors = new ObservableCollection<EditorBase>();
		private readonly SpecialEdoitorsHolder _specialEdoitorsHolder = SpecialEdoitorsHolder.Load();

		private readonly RecentFilesHolder _recentFilesHolder = RecentFilesHolder.Load();

		private EditorBase _selectedEditor;

		public ObservableCollection<EditorBase> OpenedEditors
		{
			get { return _openedEditors; }
		}

		internal string FilePath
		{
			get
			{
				if (_selectedEditor == null) 
					return null;

				if (string.IsNullOrEmpty(_selectedEditor.FilePath))
					return "New File";

				return _selectedEditor.FilePath;
			}
		}

		public RecentFilesHolder RecentFilesHolder
		{
			get { return _recentFilesHolder; }
		}

		internal IFormat[] Editors
		{
			get { return _editors; }
		}

		public EditorBase OpenFile(string fileName, bool isNew, bool needSelect)
		{
			return OpenFile(fileName, OpenedEditors.Count, isNew, needSelect, true);
		}

		public EditorBase OpenFile(string fileName, bool isNew, bool needSelect, bool needRegisterAsRecentOpened)
		{
			return OpenFile(fileName, OpenedEditors.Count, isNew, needSelect, needRegisterAsRecentOpened);
		}

		public EditorBase OpenFile(string fileName, int indexInTabOrder, bool isNew, bool needSelect, bool needRegisterAsRecentOpened)
		{
			EditorBase existedEditor = _openedEditors.FirstOrDefault(x => x.FilePath != null && x.FilePath.Equals(fileName, StringComparison.InvariantCultureIgnoreCase));

			if (existedEditor != null && !isNew)
			{
				SelectedEditor = existedEditor;
				_recentFilesHolder.RegisterPath(fileName);
				return existedEditor;
			}

			IFormat format;
			if (_specialEdoitorsHolder.ContainsInfo(fileName))
			{
				byte id = _specialEdoitorsHolder.GetInfo(fileName);
				format = _editors.First(x => x.EditorTypeId == id);
			}
			else
			{
				string fileExtension = Path.GetExtension(fileName).ToLower();
				format = _editors.FirstOrDefault(x => x.Extension.Split('|').Contains(fileExtension)) ?? _editors.First();
			}
			
			FileWrapper file = new FileWrapper(fileName, isNew);
			EditorBase openedEditor = format.CreateNewEditor(this, file);

			OpenedEditors.Insert(indexInTabOrder, openedEditor);

			if (needSelect)
				SelectedEditor = openedEditor;

			if (!isNew && needRegisterAsRecentOpened)
				_recentFilesHolder.RegisterPath(fileName);

			return openedEditor;
		}

		internal void CreateNewFile()
		{
			EditorBase openedEditor = new PlainTextEditor(this, new FileWrapper(string.Empty, true));

			OpenedEditors.Add(openedEditor);
			SelectedEditor = openedEditor;
		}

		internal EditorBase SelectedEditor
		{
			get { return _selectedEditor; }
			set
			{
				if (_selectedEditor == value)
					return;

				_selectedEditor = value;

				if (SelectedEditorChanged != null)
					SelectedEditorChanged();
			}
		}

		internal void SaveFile()
		{
			_selectedEditor.SaveFile();
		}

		internal void SaveAll()
		{
			_openedEditors.ToList().ForEach(x =>
			{
				if (!string.IsNullOrEmpty(x.FilePath))
					x.SaveFile();
			});

			if (AllHeaderChanged != null)
				AllHeaderChanged();
		}

		internal void SaveFileAs(string fileName)
		{
			_selectedEditor.SaveFileAs(fileName);
			_recentFilesHolder.RegisterPath(fileName);

			if (HeaderChanged != null)
				HeaderChanged();
		}

		internal void SaveCopyFileAs(string fileName)
		{
			_selectedEditor.SaveCopyAs(fileName);
		}

		internal void CloseActiveEditor()
		{
			EditorBase editor = SelectedEditor;
			SelectedEditor = null;

			if (!editor.IsNew)
				_recentFilesHolder.RegisterPath(editor.FilePath);

			editor.FileWrapper.Dispose();
			editor.Dispose();
			OpenedEditors.Remove(editor);
		}

		public void RefreshActiveTitle()
		{
			if (HeaderChanged != null)
				HeaderChanged();
		}

		internal void CloseAllEditors()
		{
			while (OpenedEditors.Count > 0)
			{
				EditorBase editor = OpenedEditors.First();

				if (!editor.IsNew)
					_recentFilesHolder.RegisterPath(editor.FilePath);

				editor.FileWrapper.Dispose();
				editor.Dispose();
				OpenedEditors.Remove(editor);
			}

			SelectedEditor = null;
		}

		internal void CloseAllEditorsButThis()
		{
			CloseAllEditorsButThis(_selectedEditor);
		}

		internal void CloseAllEditorsButThis(EditorBase target)
		{
			for (int index = _openedEditors.Count - 1; index >= 0; index--)
			{
				EditorBase editor = _openedEditors[index];

				if (editor == target)
					continue;

				if (!editor.IsNew)
					_recentFilesHolder.RegisterPath(editor.FilePath);

				editor.FileWrapper.Dispose();
				editor.Dispose();
				_openedEditors.Remove(editor);
			}
		}

		internal void CloseSpecifiedEditor(EditorBase editor)
		{
			if (SelectedEditor == editor)
				SelectedEditor = null;

			if (!editor.IsNew)
				_recentFilesHolder.RegisterPath(editor.FilePath);

			editor.FileWrapper.Dispose();
			editor.Dispose();
			OpenedEditors.Remove(editor);
		}

		internal void SelectEditorCommand(EditorBase arg)
		{
			SelectedEditor = arg;
		}

		internal void ChangeContentType(IFormat arg)
		{
			string fileExtension = Path.GetExtension(_selectedEditor.FileWrapper.FilePath);
			bool isDefaultEditor = arg.Extension.Split('|').Any(x => x.Equals(fileExtension, StringComparison.InvariantCultureIgnoreCase));

			if (isDefaultEditor)
				_specialEdoitorsHolder.UnregisterPath(_selectedEditor.FileWrapper.FilePath);
			else
				_specialEdoitorsHolder.RegisterPath(_selectedEditor.FileWrapper.FilePath, arg.EditorTypeId);

			EditorBase openedEditor = arg.CreateNewEditor(this, _selectedEditor.FileWrapper);

			_selectedEditor.Dispose();
			OpenedEditors[OpenedEditors.IndexOf(_selectedEditor)] = openedEditor;

			_selectedEditor = openedEditor;
		}

		internal void Reload()
		{
			_selectedEditor.Reload();
		}

		void IEditorOwner.CloseEditor(EditorBase target)
		{
			CloseSpecifiedEditor(target);
		}

		void IEditorOwner.CloseActiveEditor()
		{
			CloseActiveEditor();
		}
	}
}
