﻿using Microsoft.Win32;
using Raccoom.Xml;
using System;
using System.Collections.ObjectModel;
using System.Diagnostics;
using System.IO;
using System.Reflection;
using System.Windows;
using System.Windows.Input;
using UniversalEditor.Base;
using UniversalEditor.Base.Mvvm;
using UniversalEditor.Base.Utils.LoggerModule;
using UniversalEditor.Formats;
using UniversalEditor.OptionsModule;

namespace UniversalEditor
{
	public class DomainViewModel : ViewModelBase
	{
		private readonly DomainModel _model;

		private readonly SimpleCommand _saveCommand;
		private readonly SimpleCommand _saveAllCommand;
		private readonly SimpleCommand _saveAsCommand;
		private readonly SimpleCommand _saveCopyAsCommand;
		private readonly SimpleCommand _reloadCommand;
		private readonly SimpleCommand _closeActiveCommand;
		private readonly SimpleCommand<EditorBase> _closeCommand;
		private readonly SimpleCommand _closeAllCommand;
		private readonly SimpleCommand<EditorBase> _closeAllButThisCommand;
		private readonly SimpleCommand _closeAllButActiveCommand;

		internal DomainViewModel(DomainModel model)
		{
			if (model == null) 
				throw new ArgumentNullException("model");

			_model = model;
			_model.SelectedEditorChanged += OnSelectedEditorChanged;
			_model.HeaderChanged += OnHeaderChanged;
			_model.OpenedEditors.CollectionChanged += (sender, args) => _closeAllButThisCommand.RiseCanExecuteChanged();

			_saveCommand = new SimpleCommand(OnSaveFile, () => HasOpenedEditor);
			_saveAllCommand = new SimpleCommand(_model.SaveAll, () => HasOpenedEditor);
			_saveAsCommand = new SimpleCommand(OnSaveAsFile, () => HasOpenedEditor);
			_saveCopyAsCommand = new SimpleCommand(OnSaveCopyAsFile, () => HasOpenedEditor);
			_reloadCommand = new SimpleCommand(_model.Reload, () => HasOpenedEditor);
			_closeActiveCommand = new SimpleCommand(CloseActiveEditor, () => HasOpenedEditor);
			_closeCommand = new SimpleCommand<EditorBase>(CloseEditor, editor => editor != null);
			_closeAllCommand = new SimpleCommand(CloseAllEditors, () => HasOpenedEditor);
			_closeAllButThisCommand = new SimpleCommand<EditorBase>(CloseAllEditorsButThis, editor => _model.OpenedEditors.Count > 1);
			_closeAllButActiveCommand = new SimpleCommand(CloseAllEditorsButActive, () => _model.OpenedEditors.Count > 1);
		}

		public ObservableCollection<string> RecentFiles
		{
			get { return _model.RecentFilesHolder.Files; }
		}

		private void CloseEditor(EditorBase editor)
		{
			_model.CloseSpecifiedEditor(editor);
			_closeAllButThisCommand.RiseCanExecuteChanged();
		}

		private void CloseActiveEditor()
		{
			_model.CloseActiveEditor();
			_closeAllButThisCommand.RiseCanExecuteChanged();
		}

		private void CloseAllEditors()
		{
			_model.CloseAllEditors();
			_closeAllButThisCommand.RiseCanExecuteChanged();
		}

		private void CloseAllEditorsButThis(EditorBase editor)
		{
			_model.CloseAllEditorsButThis(editor);

			_closeAllButThisCommand.RiseCanExecuteChanged();
			_closeAllButActiveCommand.RiseCanExecuteChanged();
		}

		private void CloseAllEditorsButActive()
		{
			_model.CloseAllEditorsButThis();

			_closeAllButThisCommand.RiseCanExecuteChanged();
			_closeAllButActiveCommand.RiseCanExecuteChanged();
		}

		private void OnHeaderChanged()
		{
			RaisePropertyChanged("WindowTitle");
		}

		private void OnSelectedEditorChanged()
		{
			_saveCommand.RiseCanExecuteChanged();
			_saveAllCommand.RiseCanExecuteChanged();
			_saveAsCommand.RiseCanExecuteChanged();
			_reloadCommand.RiseCanExecuteChanged();
			_closeCommand.RiseCanExecuteChanged();
			_closeAllCommand.RiseCanExecuteChanged();

			RaisePropertyChanged("WindowTitle");
			RaisePropertyChanged("SelectedEditor");
			RaisePropertyChanged("HasOpenedEditor");
		}

		public ObservableCollection<EditorBase> OpenedEditors
		{
			get { return _model.OpenedEditors; }
		}

		internal DomainModel Model
		{
			get { return _model; }
		}

		public string WindowTitle
		{
			get { return string.IsNullOrEmpty(_model.FilePath) ? "Universal File Editor" : _model.FilePath; }
		}

		public IFormat[] Editors
		{
			get { return _model.Editors; }
		}

		public ICommand OpenCommand
		{
			get {return new SimpleCommand(OnOpenFile);}
		}

		public ICommand NewCommand
		{
			get {return new SimpleCommand(CreateNewFile);}
		}
		
		public ICommand SaveCommand
		{
			get { return _saveCommand; }
		}

		public ICommand SaveAllCommand
		{
			get { return _saveAllCommand; }
		}

		public ICommand RealoadCommand
		{
			get { return _reloadCommand; }
		}

		public ICommand SaveAsCommand
		{
			get { return _saveAsCommand; }
		}

		public ICommand SaveCopyAsCommand
		{
			get { return _saveCopyAsCommand; }
		}

		public ICommand CloseCommand
		{
			get { return _closeCommand; }
		}

		public ICommand CloseActiveCommand
		{
			get { return _closeActiveCommand; }
		}

		public ICommand CopyFileNameCommand
		{
			get { return new SimpleCommand(CopyFileName); }
		}

		public ICommand CopyFilePathCommand
		{
			get { return new SimpleCommand(CopyFilePath); }
		}

		public ICommand OpenFileLocationCommand
		{
			get { return new SimpleCommand(OpenFileLocation); }
		}

		public ICommand CloseSpecifiedEditorCommand
		{
			get { return new SimpleCommand<EditorBase>(_model.CloseSpecifiedEditor); }
		}

		public ICommand CloseAllCommand
		{
			get { return _closeAllCommand; }
		}

		public ICommand OptionsCommand
		{
			get { return new SimpleCommand(OnOptions); }
		}

		private void OnOptions()
		{
			OptionsView view = new OptionsView();
			view.Owner = Application.Current.MainWindow;
			view.ShowDialog();
		}

		public ICommand CloseAllButThisCommand
		{
			get { return _closeAllButThisCommand; }
		}

		public ICommand CloseAllButActiveCommand
		{
			get { return _closeAllButActiveCommand; }
		}

		public ICommand AboutCommand
		{
			get { return new SimpleCommand(() => new AboutBox().ShowDialog()); }
		}

		public ICommand UpdateCommand
		{
			get { return new SimpleCommand(OnCheckUpdate); }
		}

		public ICommand DocumentationCommand
		{
			get { return new SimpleCommand(OnDocumentation); }
		}

		private void OnDocumentation()
		{
			Process.Start("http://universalfileeditor.codeplex.com/documentation");
		}

		private void OnCheckUpdate()
		{
			try
			{
				Uri uri = new Uri("http://universalfileeditor.codeplex.com/project/feeds/rss?ProjectRSSFeed=codeplex%3a%2f%2frelease%2funiversalfileeditor");
				RssChannel myRssChannel = new RssChannel(uri);

				RssItem item = myRssChannel.Items[0];
				string title = item.Title;
				string appKey = "Universal File Editor";


				string appName = title.Substring(title.IndexOf(appKey));

				string version = appName.Substring(appKey.Length);
				version = version.Substring(0, version.IndexOf("(")).Trim();

				if (version == Assembly.GetExecutingAssembly().GetName().Version.ToString())
				{
					MessageBox.Show(App.Current.MainWindow, "You have the latest version of this program.", "Update", MessageBoxButton.OK, MessageBoxImage.Information);
				}
				else
				{
					Process.Start(item.Link);
				}
			}
			catch (Exception exception)
			{
				Logger.Log(exception);
			}
		}

		public ICommand ContentTypeChangeCommand
		{
			get { return new SimpleCommand<IFormat>(_model.ChangeContentType); }
		}

		public ICommand OpenSpecifiedFileCommand
		{
			get { return new SimpleCommand<string>(OnOpenSpecifiedFile); }
		}

		private void OnOpenSpecifiedFile(string path)
		{
			if (!File.Exists(path))
			{
				MessageBox.Show(Application.Current.MainWindow, string.Format("File not found: '{0}'", path));
				return;
			}

			_model.OpenFile(path, false, true);
		}

		public ICommand SelectEditorCommand
		{
			get { return new SimpleCommand<EditorBase>(_model.SelectEditorCommand); }
		}

		private void CopyFileName()
		{
			Clipboard.SetText(Path.GetFileName(_model.FilePath));
		}

		private void CopyFilePath()
		{
			Clipboard.SetText(_model.FilePath);
		}

		private void OpenFileLocation()
		{
			if (string.IsNullOrEmpty(_model.FilePath))
				return;

			Process.Start("explorer.exe", @"/select, " + _model.FilePath);
		}

		public EditorBase SelectedEditor
		{
			get { return _model.SelectedEditor; }
			set
			{
				_model.SelectedEditor = value;
			}
		}

		public bool HasOpenedEditor
		{
			get { return _model.SelectedEditor != null; }
		}

		private void OnSaveFile()
		{
			if (string.IsNullOrEmpty(_model.SelectedEditor.FilePath))
				OnSaveAsFile();
			else
				_model.SaveFile();
		}

		private void OnSaveAsFile()
		{
			SaveFileDialog dialog = new SaveFileDialog();
			bool? result = dialog.ShowDialog(Application.Current.MainWindow);

			if (!result.Value)
				return;

			_model.SaveFileAs(dialog.FileName);
		}

		private void OnSaveCopyAsFile()
		{
			SaveFileDialog dialog = new SaveFileDialog();
			bool? result = dialog.ShowDialog(Application.Current.MainWindow);

			if (!result.Value)
				return;

			_model.SaveCopyFileAs(dialog.FileName);
		}

		private void OnOpenFile()
		{
			string defaultPath = _model.SelectedEditor != null && !_model.SelectedEditor.IsNew ? Path.GetDirectoryName(_model.SelectedEditor.FilePath) : string.Empty;

			OpenFileDialog dialog = new OpenFileDialog();

			if (!string.IsNullOrEmpty(defaultPath))
				dialog.InitialDirectory = defaultPath;

			bool? result = dialog.ShowDialog(Application.Current.MainWindow);

			if (!result.Value)
				return;

			_model.OpenFile(dialog.FileName, false, true);

			_closeAllButThisCommand.RiseCanExecuteChanged();
		}

		private void CreateNewFile()
		{
			_model.CreateNewFile();
			_closeAllButThisCommand.RiseCanExecuteChanged();
		}
	}
}
