﻿using System.Windows.Input;
using Raccoom.Xml;
using System;
using System.Collections.ObjectModel;
using System.Collections.Specialized;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Threading;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Xml.Linq;
using UniversalEditor.Base;
using UniversalEditor.Base.Utils.LoggerModule;

namespace UniversalEditor
{
	/// <summary>
	/// Interaction logic for MainWindow.xaml
	/// </summary>
	public partial class MainWindow
	{
		private const string WindowPositionFileName = "window.position.xml";
		private const string DocumentListFileName = "opened.documents.xml";

		public MainWindow()
		{
			DataContextChanged += OnDataContextChanged;
			InitializeComponent();
		}
		
		private void OnDataContextChanged(object sender, DependencyPropertyChangedEventArgs e)
		{
			DomainViewModel viewModel = (DomainViewModel) DataContext;
			viewModel.OpenedEditors.CollectionChanged += OnCollectionChanged;
			viewModel.Model.SelectedEditorChanged += OnModelSelectedEditorChanged;
			viewModel.Model.HeaderChanged += OnHeaderChanged;
			viewModel.Model.AllHeaderChanged += OnAllHeaderChanged;
		}

		private void OnHeaderChanged()
		{
			EditorBase tabSelected = (EditorBase)((TabItem)tabControl.SelectedItem).Tag;
			TextBlock block = (TextBlock)((StackPanel)((TabItem)tabControl.SelectedItem).Header).Children[0];

			block.Text = tabSelected.FileName;
		}

		private void OnAllHeaderChanged()
		{
			foreach (var tab in tabControl.Items.OfType<TabItem>().Select(x => new { Tab = x, Editor = (EditorBase)x.Tag }))
			{
				TextBlock block = (TextBlock)((StackPanel)(tab.Tab).Header).Children[0];
				block.Text = tab.Editor.FileName;
			}
		}

		private void OnModelSelectedEditorChanged()
		{
			DomainViewModel viewModel = (DomainViewModel)DataContext;
			EditorBase tabSelected = tabControl.SelectedItem == null ? null : (EditorBase)((TabItem)tabControl.SelectedItem).Tag;

			if (viewModel.SelectedEditor == tabSelected)
				return;

			if (viewModel.SelectedEditor == null)
			{
				tabControl.SelectedItem = tabControl.Items.OfType<TabItem>().LastOrDefault();
				return;
			}

			tabControl.SelectedItem = tabControl.Items.OfType<TabItem>().First(x => x.Tag == viewModel.SelectedEditor);
		}
		
		private void OnCollectionChanged(object sender, NotifyCollectionChangedEventArgs e)
		{
			DomainViewModel viewModel = (DomainViewModel)DataContext;
			ObservableCollection<EditorBase> collection = (ObservableCollection<EditorBase>)sender;

			if (e.Action == NotifyCollectionChangedAction.Replace)
			{
				EditorBase oldEditor = (EditorBase)e.OldItems[0];
				EditorBase newEditor = (EditorBase)e.NewItems[0];

				TabItem tabItem = tabControl.Items.OfType<TabItem>().First(x => x.Tag == oldEditor);
				tabItem.Tag = newEditor;

				if (collection.Count > 1)
				{
					tabItem.Content = newEditor.EditorControl;
				}
				else
				{
					border.Child = newEditor.EditorControl;
				}

				return;
			}

			if (e.OldItems != null)
			{
				// remove
				EditorBase[] editors = e.OldItems.OfType<EditorBase>().ToArray();

				for (int index = 0; index < editors.Length; index++)
				{
					EditorBase editor = editors[index];
					tabControl.Items.RemoveAt(tabControl.Items.OfType<TabItem>().ToList().FindIndex(x => x.Tag == editor));
				}
			}

			if (e.NewItems != null)
			{
				// add
				EditorBase[] editors = e.NewItems.OfType<EditorBase>().ToArray();

				for (int index = 0; index < editors.Length; index++)
				{
					EditorBase editor = editors[index];
					TabItem tabItem = new TabItem();
					tabItem.Tag = editor;

					TextBlock headerText = new TextBlock();
					headerText.Text = editor.FileName;
					headerText.Name = "txtHeader";
					headerText.VerticalAlignment = VerticalAlignment.Center;
					headerText.ContextMenu = new ContextMenu();
					headerText.ContextMenu.Items.Add(new MenuItem { Header = "Reload", Command = viewModel.RealoadCommand });
					headerText.ContextMenu.Items.Add(new Separator());
					headerText.ContextMenu.Items.Add(new MenuItem { Header = "Save", Command = viewModel.SaveCommand });
					headerText.ContextMenu.Items.Add(new MenuItem { Header = "Save As", Command = viewModel.SaveAsCommand });
					headerText.ContextMenu.Items.Add(new MenuItem { Header = "Save Copy As", Command = viewModel.SaveCopyAsCommand });
					headerText.ContextMenu.Items.Add(new MenuItem { Header = "Save All", Command = viewModel.SaveAllCommand });
					headerText.ContextMenu.Items.Add(new Separator());
					headerText.ContextMenu.Items.Add(new MenuItem { Header = "Close", Command = viewModel.CloseCommand, CommandParameter = editor });
					headerText.ContextMenu.Items.Add(new MenuItem { Header = "Close All", Command = viewModel.CloseAllCommand });
					headerText.ContextMenu.Items.Add(new MenuItem { Header = "Close All But This", Command = viewModel.CloseAllButThisCommand, CommandParameter = editor });
					headerText.ContextMenu.Items.Add(new Separator());
					headerText.ContextMenu.Items.Add(new MenuItem { Header = "Copy File Name", Command = viewModel.CopyFileNameCommand });
					headerText.ContextMenu.Items.Add(new MenuItem { Header = "Copy File Path", Command = viewModel.CopyFilePathCommand });
					headerText.ContextMenu.Items.Add(new MenuItem { Header = "Open File Location", Command = viewModel.OpenFileLocationCommand });

					Button closeTabButton = new Button();
					closeTabButton.Command = viewModel.CloseCommand;
					closeTabButton.CommandParameter = editor;
					closeTabButton.VerticalAlignment = VerticalAlignment.Center;
					closeTabButton.VerticalContentAlignment = VerticalAlignment.Center;
					closeTabButton.Margin = new Thickness(5, 0, 0, 0);
					closeTabButton.Template = (ControlTemplate) Resources["btnCloseTemplate"];
					
					StackPanel header = new StackPanel();
					header.Orientation = Orientation.Horizontal;
					header.FocusVisualStyle = null;
					header.Focusable = false;
					header.SetBinding(ToolTipProperty, new Binding {Source = editor, Path = new PropertyPath("FilePath")});
					header.Children.Add(headerText);
					header.Children.Add(closeTabButton);

					tabItem.Header = header;
					tabItem.FocusVisualStyle = null;
					tabControl.Items.Insert(e.NewStartingIndex + index, tabItem);

					tabItem.UpdateLayout();
//					editor.Focus();
				}
			}

			if (tabControl.SelectedItem == null)
				tabControl.SelectedItem = tabControl.Items.OfType<TabItem>().LastOrDefault();

			if (collection.Count == 0)
			{
				tabControl.Visibility = Visibility.Collapsed;
				border.Visibility = Visibility.Collapsed;
				border.Child = null;
				return;
			}

			TabItem tab = (TabItem) tabControl.SelectedItem;
			EditorBase selectedEditor = (EditorBase)tab.Tag;
			if (collection.Count == 1 && border.Visibility != Visibility.Visible)
			{
				tabControl.Visibility = Visibility.Collapsed;
				
				tab.Content = null;
				border.Child = selectedEditor.EditorControl;

				border.Visibility = Visibility.Visible;
				border.UpdateLayout();
				collection.First().Focus();
				return;
			}

			if (collection.Count > 1 && border.Visibility == Visibility.Visible)
			{
				border.Visibility = Visibility.Collapsed;
				border.Child = null;

				tab.Content = selectedEditor.EditorControl;

				tabControl.Visibility = Visibility.Visible;
				return;
			}
		}

		protected override void OnKeyDown(KeyEventArgs e)
		{
			if (!e.Handled)
			{
				DomainViewModel viewModel = (DomainViewModel) DataContext;
				if (viewModel.SelectedEditor != null)
				{
					viewModel.SelectedEditor.OnKeyDown(e);
				}
			}

			base.OnKeyDown(e);
		}

		protected override void OnTextInput(TextCompositionEventArgs e)
		{
			if (!e.Handled)
			{
				DomainViewModel viewModel = (DomainViewModel)DataContext;
				if (viewModel.SelectedEditor != null)
				{
					viewModel.SelectedEditor.OnTextInput(e);
				}
			}

			base.OnTextInput(e);
		}

		protected override void OnInitialized(EventArgs e)
		{
			LoadWindowPosition();

			base.OnInitialized(e);

			LoadOpenedDocumentsList();

			if (!string.IsNullOrEmpty(((App) Application.Current).StartFilePath))
				App.DomainModel.Model.OpenFile(((App)Application.Current).StartFilePath, false, true);

			Thread updateThread = new Thread(() =>
			{
				try
				{
					const string appKey = "Universal File Editor";

					Uri uri = new Uri("http://universalfileeditor.codeplex.com/project/feeds/rss?ProjectRSSFeed=codeplex%3a%2f%2frelease%2funiversalfileeditor");
					RssChannel myRssChannel = new RssChannel(uri);

					for (int index = 0; index < myRssChannel.Items.Count; index++)
					{
						RssItem item = myRssChannel.Items[index];
						string title = item.Title;

						if (!title.Contains(appKey + " 0.3"))
							continue;

						string appName = title.Substring(title.IndexOf(appKey));
						string version = appName.Substring(appKey.Length);
						version = version.Substring(0, version.IndexOf("(")).Trim();

						if (version == Assembly.GetExecutingAssembly().GetName().Version.ToString())
							return;

						Application.Current.Dispatcher.Invoke(() =>
						{
							MessageBoxResult result = MessageBox.Show(Application.Current.MainWindow, "The newer version of this program is available. Do you want to download it?", "Update",
																	  MessageBoxButton.OKCancel, MessageBoxImage.Information);

							if (result == MessageBoxResult.OK)
								Process.Start(item.Link);
						});
					}
				}
				catch { }
			});

			updateThread.IsBackground = true;
			updateThread.SetApartmentState(ApartmentState.STA);
			updateThread.Start();
		}

		protected override void OnClosing(System.ComponentModel.CancelEventArgs e)
		{
			bool isSuccess = SaveEditedFiles();

			if (!isSuccess)
			{
				e.Cancel = true;
				return;
			}

			SaveWindowPosition();
			SaveOpenedDocumentsList();

			base.OnClosing(e);
		}

		private bool SaveEditedFiles()
		{
			DomainViewModel viewModel = (DomainViewModel)DataContext;

			for (int index = 0; index < viewModel.OpenedEditors.Count; index++)
			{
				EditorBase editor = viewModel.OpenedEditors[index];

				if (string.IsNullOrEmpty(editor.FilePath))
					continue;

				if (!editor.HasChanges)
					continue;

				MessageBoxResult result = MessageBox.Show(this, string.Format("File '{0}' has been changed. Do you want to save it?", editor.FilePath),
				                "File has been changed", MessageBoxButton.YesNoCancel, MessageBoxImage.Warning);

				if (result == MessageBoxResult.Yes)
					editor.SaveFile();
				else if (result == MessageBoxResult.No)
					continue;
				else
					return false;
			}

			return true;
		}

		private void TabControl_OnSelectionChanged(object sender, SelectionChangedEventArgs e)
		{
			if (tabControl.SelectedItem == null)
				return;

			TabItem tabItem = (TabItem) tabControl.SelectedItem;
			EditorBase editorBase = (EditorBase)tabItem.Tag;

			DomainViewModel viewModel = (DomainViewModel)DataContext;

			tabItem.Content = editorBase.EditorControl;
			viewModel.SelectedEditor = editorBase;
			
			viewModel.SelectedEditor.Focus();
		}

		private void SaveWindowPosition()
		{
			XDocument doc = new XDocument(
				new XElement("Window",
					new XElement("Left", Left),
					new XElement("Top", Top),
					new XElement("Height", Height),
					new XElement("Width", Width),
					new XElement("Maximize", WindowState == WindowState.Maximized)));

			string filePath = Path.Combine(Path.Combine(Environment.GetFolderPath(Environment.SpecialFolder.UserProfile), "UniversalFileEditor"), WindowPositionFileName);

			if (!Directory.Exists(Path.GetDirectoryName(filePath)))
				Directory.CreateDirectory(Path.GetDirectoryName(filePath));

			doc.Save(filePath);
		}

		private void LoadWindowPosition()
		{
			string filePath = Path.Combine(Path.Combine(Environment.GetFolderPath(Environment.SpecialFolder.UserProfile), "UniversalFileEditor"), WindowPositionFileName);

			if (!File.Exists(filePath))
				return;

			XDocument doc = XDocument.Load(filePath);

			WindowState = Convert.ToBoolean(doc.Root.Element("Maximize").Value) ? WindowState.Maximized : WindowState.Normal;
			Left = Convert.ToDouble(doc.Root.Element("Left").Value);
			Top = Convert.ToDouble(doc.Root.Element("Top").Value);
			Height = Convert.ToDouble(doc.Root.Element("Height").Value);
			Width = Convert.ToDouble(doc.Root.Element("Width").Value);
		}

		private void SaveOpenedDocumentsList()
		{
			string appDir = Path.Combine(Environment.GetFolderPath(Environment.SpecialFolder.UserProfile), "UniversalFileEditor");
			string newFilesHolderDir = Path.Combine(appDir, "NewFiles");

			if (Directory.Exists(newFilesHolderDir))
				Directory.Delete(newFilesHolderDir, true);

			DomainViewModel viewModel = (DomainViewModel)DataContext;
			XDocument doc = new XDocument(new XElement("Documents"));

			for (int index = 0; index < viewModel.OpenedEditors.Count; index++)
			{
				EditorBase editor = viewModel.OpenedEditors[index];
				
				XElement file = new XElement("Document");
				file.Add(new XElement("Path", editor.FilePath));
				file.Add(new XElement("IsSelected", editor == viewModel.SelectedEditor));

				bool isNewFile = string.IsNullOrEmpty(editor.FilePath);

				file.Add(new XElement("IsNew", isNewFile));

				if (isNewFile)
				{
					if (!Directory.Exists(newFilesHolderDir))
						Directory.CreateDirectory(newFilesHolderDir);

					string tempFilePath = Path.Combine(newFilesHolderDir, Path.GetRandomFileName());
					tempFilePath = Path.ChangeExtension(tempFilePath, ".txt");

					editor.SaveCopyAs(tempFilePath);
					file.Add(new XElement("NewFilePath", tempFilePath));
				}

				doc.Root.Add(file);
			}


			string filePath = Path.Combine(appDir, DocumentListFileName);

			if (!Directory.Exists(Path.GetDirectoryName(filePath)))
				Directory.CreateDirectory(Path.GetDirectoryName(filePath));

			doc.Save(filePath);
		}

		private void LoadOpenedDocumentsList()
		{
			string filePath = Path.Combine(Path.Combine(Environment.GetFolderPath(Environment.SpecialFolder.UserProfile), "UniversalFileEditor"), DocumentListFileName);

			if (!File.Exists(filePath))
				return;

			XDocument doc = XDocument.Load(filePath);

			if (doc.Root.Element("document") != null)
			{
				foreach (XElement docTag in doc.Root.Elements("document"))
				{
					string documentPath = docTag.Value;

					if (string.IsNullOrEmpty(documentPath) || !File.Exists(documentPath))
						continue;

					App.DomainModel.Model.OpenFile(documentPath, false, false);
				}
			}
			else
			{
				var files = (from docTag in doc.Root.Elements("Document")
							let isNew = bool.Parse(docTag.Element("IsNew").Value)
							select new
							   {
								   Path = docTag.Element("Path").Value,
								   IsNew = isNew,
								   IsSelected = bool.Parse(docTag.Element("IsSelected").Value),
								   NewFilePath = isNew ? docTag.Element("NewFilePath").Value : string.Empty
							   })
							.ToArray();

				int selectedEditorIndex = -1;

				for (int index = 0; index < files.Length; index++)
				{
					var fileItem = files[index];

					try
					{
						if (!fileItem.IsSelected)
							continue;

						string path = fileItem.IsNew ? fileItem.NewFilePath : fileItem.Path;

						if (string.IsNullOrEmpty(path) || !File.Exists(path))
							break;

						selectedEditorIndex = index;
						App.DomainModel.Model.OpenFile(path, fileItem.IsNew, true, false);

						break;
					}
					catch (Exception)
					{
//						Logger.Log(exception);
					}
				}

				for (int index = 0; index < files.Length; index++)
				{
					if (index == selectedEditorIndex)
						continue;

					var fileItem = files[index];

					try
					{
						string path = fileItem.IsNew ? fileItem.NewFilePath : fileItem.Path;

						if (string.IsNullOrEmpty(path) || !File.Exists(path))
							continue;

						EditorBase editor;

						if (index < selectedEditorIndex)
							editor = App.DomainModel.Model.OpenFile(path, App.DomainModel.Model.OpenedEditors.Count - 1, fileItem.IsNew, false, false);
						else
							editor = App.DomainModel.Model.OpenFile(path, App.DomainModel.Model.OpenedEditors.Count, fileItem.IsNew, false, false);

						if (fileItem.IsNew)
							editor.EditorControl.ToString();
					}
					catch (Exception)
					{
//						Logger.Log(exception);
					}
				}
			}
		}

		private void MainWindow_OnDragEnter(object sender, DragEventArgs e)
		{
			bool hasData = e.Data.GetDataPresent("FileDrop");

			if (!hasData)
			{
				e.Effects = DragDropEffects.None;
				e.Handled = true;
			}
		}

		private void MainWindow_OnDrop(object sender, DragEventArgs e)
		{
			string[] filesPathes = (string[])e.Data.GetData("FileDrop");

			for (int index = 0; index < filesPathes.Length; index++)
				App.DomainModel.Model.OpenFile(filesPathes[index], false, true, true);
		}
	}
}
