﻿using System;
using System.Globalization;
using System.Linq;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Controls.Primitives;
using System.Windows.Media;
using System.Windows.Threading;
using TextEditor.TextBoxInner.Layers;
using TextEditor.TextBoxInner.Utils;
using UniversalEditor.Base;
using UniversalEditor.Base.Options;

namespace TextEditor.TextBoxInner
{
	public class UfeTextBox2 : FrameworkElement, IScrollInfo
	{
		private readonly IEditorOwner _owner;
		private readonly UfeTextDocument _document;

		private readonly VisualCollection _layers;

		private readonly TextLayer _textLayer = new TextLayer();
		private readonly NumberLayer _numbersLayer = new NumberLayer();
		private readonly CollapseLayer _collapseLayer = new CollapseLayer();

		private bool _canVerticallyScroll = true;
		private bool _canHorizontallyScroll = true;

		private bool _showNumbers = Options.Instance.ShowLineNumbers;

		private Size _viewport;
		private Size _extent;
		private Vector _offset;

		private CharPosition _startSelection = new CharPosition(0, 0);
		private CharPosition _endSelection = new CharPosition(0, 0);

		public UfeTextBox2(IEditorOwner owner, UfeTextDocument document)
		{
			if (owner == null) 
				throw new ArgumentNullException("owner");
			if (document == null)
				throw new ArgumentNullException("document");

			_owner = owner;
			_document = document;

			HorizontalAlignment = HorizontalAlignment.Left;
			VerticalAlignment = VerticalAlignment.Top;

			_layers = new VisualCollection(this);
			_layers.Add(_numbersLayer);
			_layers.Add(_textLayer);
			_layers.Add(_collapseLayer);

			if (!_showNumbers)
			{
				_numbersLayer.Visibility = Visibility.Collapsed;
				_collapseLayer.Visibility = Visibility.Collapsed;
			}

			UpdateExtendSize();
		}

		private void UpdateExtendSize()
		{
			double longestLineWidth = 0;

			for (int index = 0; index < _document.LineCount; index++)
			{
				double lineWidth = _document.GetLineWidthInPixels(index);

				if (longestLineWidth < lineWidth)
					longestLineWidth = lineWidth;
			}

			_extent = new Size(longestLineWidth, _document.VisibleLineCount * FontHelper.FontHeight);
		}

		protected override Size ArrangeOverride(Size finalSize)
		{
			UpdateLayers();

			return _extent;
		}

		private void UpdateLayers()
		{
			int startLineIndex = Convert.ToInt32(_offset.Y/FontHelper.FontHeight);
			int linesCount = _document.LineCount;

			if (startLineIndex >= linesCount)
			{
				_textLayer.SetLines(new RenderedInfo[0]);
				_numbersLayer.SetLines(new int[0]);
			}
			else
			{
				int endLineIndex = Convert.ToInt32((_offset.Y + _viewport.Height) / FontHelper.FontHeight);
				int end = Math.Min(endLineIndex, linesCount - 1);

				double textLayerOffset = 2;

				if (_showNumbers)
				{
					_numbersLayer.SetLines(Enumerable.Range(startLineIndex, end - startLineIndex + 1).ToArray());

					double collapseLayerOffset = end.ToString(CultureInfo.InvariantCulture).Length * FontHelper.FontWidth + 2;
					_collapseLayer.SetLeft(collapseLayerOffset);

					textLayerOffset = collapseLayerOffset + 2;
				}
 
				_textLayer.SetLeft(textLayerOffset);
				_textLayer.SetLines(Enumerable.Range(startLineIndex, end - startLineIndex + 1).Select(x => _document.GetRenderedInfo(x)).ToArray());
			}
		}

		protected void VerifyScrollData()
		{
			if (double.IsInfinity(_viewport.Width))
				_viewport.Width = _extent.Width;

			if (double.IsInfinity(_viewport.Height))
				_viewport.Height = SystemParameters.PrimaryScreenHeight;

			_offset.X = Math.Max(0, Math.Min(_offset.X, _extent.Width - _viewport.Width));
			_offset.Y = Math.Max(0, Math.Min(_offset.Y, _extent.Height - _viewport.Height));

			if (ScrollOwner != null)
				ScrollOwner.InvalidateScrollInfo();
		}

		protected override Visual GetVisualChild(int index)
		{
			return _layers[index];
		}

		protected override int VisualChildrenCount
		{
			get { return _layers.Count; }
		}

		protected override Size MeasureOverride(Size availableSize)
		{
			_viewport = availableSize;

			VerifyScrollData();
			UpdateVisibleLinesCount();

			if (ScrollOwner != null)
				ScrollOwner.InvalidateScrollInfo();

			return _viewport;
		}

		private void UpdateVisibleLinesCount()
		{
			int linesCount = Convert.ToInt32(_viewport.Height/FontHelper.FontHeight) + 2;

			_textLayer.SetLinesCount(linesCount);
			_numbersLayer.SetLinesCount(linesCount);
		}

		public void LineUp()
		{
			SetScrollOffset(_offset.X, _offset.Y - FontHelper.FontHeight);
		}

		public void LineDown()
		{
			SetScrollOffset(_offset.X, _offset.Y + FontHelper.FontHeight);
		}

		public void LineLeft()
		{
			SetScrollOffset(_offset.X - FontHelper.FontWidth, _offset.Y);
		}

		public void LineRight()
		{
			SetScrollOffset(_offset.X + FontHelper.FontWidth, _offset.Y);
		}

		public void PageUp()
		{
		}

		public void PageDown()
		{
		}

		public void PageLeft()
		{
		}

		public void PageRight()
		{
		}

		public void MouseWheelUp()
		{
			SetScrollOffset(_offset.X, _offset.Y - 3 * FontHelper.FontHeight);
		}

		public void MouseWheelDown()
		{
			SetScrollOffset(_offset.X, _offset.Y + 3 * FontHelper.FontHeight);
		}

		public void MouseWheelLeft()
		{
			SetScrollOffset(_offset.X - 3 * FontHelper.FontWidth, _offset.Y);
		}

		public void MouseWheelRight()
		{
			SetScrollOffset(_offset.X + 3 * FontHelper.FontWidth, _offset.Y);
		}

		public void SetHorizontalOffset(double offset)
		{
			SetScrollOffset(offset, _offset.Y);
		}

		public void SetVerticalOffset(double offset)
		{
			SetScrollOffset(_offset.X, offset);
		}

		private void SetScrollOffset(double x, double y)
		{
			if (x > _extent.Width - _viewport.Width) 
				x = _extent.Width - _viewport.Width;
			if (x < 0)
				x = 0;
			if (y > _extent.Height - _viewport.Height) 
				y = _extent.Height - _viewport.Height;
			if (y < 0)
				y = 0;

			if (Math.Abs(_offset.X - x) >= 10 || Math.Abs(_offset.Y - y) >= 10)
			{
				_offset.X = x;
				_offset.Y = y;

				if (ScrollOwner != null)
					ScrollOwner.InvalidateScrollInfo();

				UpdateLayers();
			}
		}

		public Rect MakeVisible(Visual visual, Rect rectangle)
		{
			return rectangle;
		}

		public bool CanVerticallyScroll
		{
			get { return _canVerticallyScroll; }
			set { _canVerticallyScroll = value; }
		}

		public bool CanHorizontallyScroll
		{
			get { return _canHorizontallyScroll; }
			set { _canHorizontallyScroll = value; }
		}

		public double ExtentWidth
		{
			get { return _extent.Width; }
		}

		public double ExtentHeight
		{
			get { return _extent.Height; }
		}

		public double ViewportWidth 
		{
			get { return _viewport.Width; }
		}

		public double ViewportHeight
		{
			get { return _viewport.Height; }
		}

		public double HorizontalOffset
		{
			get { return _offset.X; }
		}

		public double VerticalOffset
		{
			get { return _offset.Y; }
		}

		public ScrollViewer ScrollOwner { get; set; }
	}
}
