﻿using System;
using System.Collections.Generic;
using System.Globalization;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Controls.Primitives;
using System.Windows.Input;
using System.Windows.Media;
using System.Xml.Linq;
using UniversalEditor.Base;
using UniversalEditor.Base.MenuBarItems;
using UniversalEditor.Base.Mvvm;

namespace TextEditor
{
	public delegate void KeyDownDelegate(KeyEventArgs e);

	public class UfeTextBox : Grid
	{
		private readonly IEditorOwner _owner;

		public event KeyDownDelegate OnKeyDownEvent
		{
			add { _textBox.OnKeyDownEvent += value; }
			remove { _textBox.OnKeyDownEvent -= value; }
		}

		public event Action<int> LinesCountChanged
		{
			add { _textBox.LinesCountChanged += value; }
			remove { _textBox.LinesCountChanged -= value; }
		}

		private readonly ScrollBar _verticalScroll;
		private readonly ScrollBar _horizontalScroll;
		private readonly FindLineCtrl _findLineCtrl;
		private readonly  FindWordCtrl _findWordCtrl;
		private readonly  DocumentChangedCtrl _documentChangedCtrl;
		private readonly  UfeTextBoxInner _textBox;
		private readonly UfePopup _popup;
		
		private readonly ReadonlyCtrl _readonlyCtrl;
		private readonly DocumentDeletedCtrl _documentDeletedCtrl;

		private MenuListItem _menu1;
		private MenuListItem _menu2;

		public UfeTextBoxInner TextBox
		{
			get { return _textBox; }
		}

		internal IEditorOwner Owner
		{
			get { return _owner; }
		}

		public UfeTextBox(IEditorOwner owner, UfeTextDocument doc)
		{
			if (owner == null) 
				throw new ArgumentNullException("owner");

			_owner = owner;

			BeginInit();

			// background
			UserControl background = new UserControl();
			background.Background = Brushes.Transparent;
			SetColumnSpan(background, 2);
			Children.Add(background);

			// scroll
			_verticalScroll = new ScrollBar();
//			_verticalScroll.MinWidth = 10;
			_verticalScroll.Width = 17;
			_verticalScroll.Visibility = Visibility.Collapsed;
			_verticalScroll.Orientation = Orientation.Vertical;
			_verticalScroll.Margin = new Thickness(0);
			_verticalScroll.Padding = new Thickness(0);

			SetColumn(_verticalScroll, 2);
			SetRow(_verticalScroll, 0);
			SetRowSpan(_verticalScroll, 3);
			Children.Add(_verticalScroll);

			_horizontalScroll = new ScrollBar();
//			_horizontalScroll.MinHeight = 10;
			_horizontalScroll.Height = 17;
			_horizontalScroll.Visibility = Visibility.Collapsed;
			_horizontalScroll.Orientation = Orientation.Horizontal;
			_horizontalScroll.Margin = new Thickness(0);
			_horizontalScroll.Padding = new Thickness(0);

			SetColumnSpan(_horizontalScroll, 2);
			SetRow(_horizontalScroll, 3);
			Children.Add(_horizontalScroll);
			
			// find line
			_findLineCtrl = new FindLineCtrl(this);
			_findLineCtrl.Visibility = Visibility.Collapsed;
			_findLineCtrl.Margin = new Thickness(0);
			_findLineCtrl.Padding = new Thickness(0);
			SetColumn(_findLineCtrl, 1);
			SetRow(_findLineCtrl, 0);

			// find word
			_findWordCtrl = new FindWordCtrl(this);
			_findWordCtrl.Visibility = Visibility.Collapsed;
			_findWordCtrl.Margin = new Thickness(0);
			_findWordCtrl.Padding = new Thickness(0);
			SetColumn(_findWordCtrl, 1);
			SetRow(_findWordCtrl, 0);

			// readonly
			_readonlyCtrl = new ReadonlyCtrl(this);
			_readonlyCtrl.Visibility = Visibility.Collapsed;
			_readonlyCtrl.HorizontalAlignment = HorizontalAlignment.Center;
			_readonlyCtrl.Margin = new Thickness(0);
			_readonlyCtrl.Padding = new Thickness(0);
			SetColumnSpan(_readonlyCtrl, 2);
			SetRow(_readonlyCtrl, 2);

			// popup
			_popup = new UfePopup(this);
			_popup.Visibility = Visibility.Collapsed;
			_popup.Placement = PlacementMode.Custom;
			_popup.CustomPopupPlacementCallback = CustomPopupPlacementCallback;

			// file changed
			_documentChangedCtrl = new DocumentChangedCtrl(this);
			_documentChangedCtrl.Visibility = Visibility.Collapsed;
			_documentChangedCtrl.HorizontalAlignment = HorizontalAlignment.Center;
			_documentChangedCtrl.Margin = new Thickness(0);
			_documentChangedCtrl.Padding = new Thickness(0);
			SetColumnSpan(_documentChangedCtrl, 2);
			SetRow(_documentChangedCtrl, 2);

			// file changed
			_documentDeletedCtrl = new DocumentDeletedCtrl(owner, this);
			_documentDeletedCtrl.Visibility = Visibility.Collapsed;
			_documentDeletedCtrl.HorizontalAlignment = HorizontalAlignment.Center;
			_documentDeletedCtrl.Margin = new Thickness(0);
			_documentDeletedCtrl.Padding = new Thickness(0);
			SetColumnSpan(_documentDeletedCtrl, 2);
			SetRow(_documentDeletedCtrl, 2);

			// _textBox
			_textBox = new UfeTextBoxInner(this, doc, _verticalScroll, _horizontalScroll, _findLineCtrl, _findWordCtrl, _documentChangedCtrl, _documentDeletedCtrl, _readonlyCtrl, _popup);
			_textBox.ClipToBounds = true;
			_textBox.Cursor = Cursors.IBeam;
			SetRowSpan(_textBox, 3);
			SetColumnSpan(_textBox, 2);

			Children.Add(_textBox);
			Children.Add(_findLineCtrl);
			Children.Add(_findWordCtrl);
			Children.Add(_readonlyCtrl);
			Children.Add(_popup);
			Children.Add(_documentChangedCtrl);
			Children.Add(_documentDeletedCtrl);

			ColumnDefinitions.Add(new ColumnDefinition { Width = new GridLength(100, GridUnitType.Star) });
			ColumnDefinitions.Add(new ColumnDefinition { Width = new GridLength(1, GridUnitType.Auto) });
			ColumnDefinitions.Add(new ColumnDefinition { Width = new GridLength(1, GridUnitType.Auto) });

			RowDefinitions.Add(new RowDefinition { Height = new GridLength(1, GridUnitType.Auto) });
			RowDefinitions.Add(new RowDefinition { Height = new GridLength(100, GridUnitType.Star) });
			RowDefinitions.Add(new RowDefinition { Height = new GridLength(1, GridUnitType.Auto) });
			RowDefinitions.Add(new RowDefinition { Height = new GridLength(1, GridUnitType.Auto) });

			ClipToBounds = true;
			Margin = new Thickness(0);
			_popup.PlacementTarget = _textBox;
			
			if (doc.IsNewFile)
				IsReadonly = false;

			_textBox.TabIndex = 1;
			_textBox.Focus();

			EndInit();
		}

		private CustomPopupPlacement[] CustomPopupPlacementCallback(Size popupSize, Size targetSize, Point offset)
		{
			CustomPopupPlacement placement1 = new CustomPopupPlacement(_textBox.GetCursorPosition(), PopupPrimaryAxis.Vertical);
			CustomPopupPlacement placement2 = new CustomPopupPlacement(_textBox.GetCursorPosition(), PopupPrimaryAxis.Horizontal);

			CustomPopupPlacement[] ttplaces = new[] { placement1, placement2 };
			return ttplaces;
		}

		public void ShowLine(int number)
		{
			_textBox.ShowLine(number);
		}

		public UfeTextDocument Document
		{
			get { return _textBox.Document; }
		}

		public void FindText(string text, bool useRegister)
		{
			_textBox.FindText(text, useRegister);
		}

		internal int ReplaceAllText(string text, bool useRegister, string value)
		{
			return _textBox.ReplaceAllText(text, useRegister, value);
		}

		internal void ReplaceText(string text, bool useRegister, string value)
		{
			_textBox.ReplaceText(text, useRegister, value);
		}

		public void HideAllFindControls()
		{
			_textBox.SetTopPanelVisibility(UfeTextBoxInner.TopPanelVisibility.None);
		}

		public void HideAllBottomControls()
		{
			_textBox.SetBottomPanelVisibility(UfeTextBoxInner.BottomPanelVisibility.None);
		}

		public bool IsReadonly
		{
			get { return _textBox.IsReadonly; }
			set
			{
				if (_textBox.IsReadonly == value)
					return;

				_textBox.IsReadonly = value;
				HideAllBottomControls();
			}
		}

		public void CallOnRender()
		{
			_textBox.InvalidateVisual();
		}

		public MenuListItem GenerateBasicMenu()
		{
			if (_menu1 == null)
			{
				List<MenuItemBase> subItems = new List<MenuItemBase>();
				subItems.Add(new MenuCommandItem("Find...", null, "Ctrl+F", new SimpleCommand(_textBox.OnFind)));
				subItems.Add(new MenuCommandItem("Replace...", null, "Ctrl+H", new SimpleCommand(_textBox.OnReplace)));
				subItems.Add(new MenuCommandItem("Go To...", null, "Ctrl+G", new SimpleCommand(_textBox.OnFindLine)));
				subItems.Add(new MenuSeparator());
				subItems.Add(new MenuCommandItem("Duplicate...", null, "Ctrl+D", new SimpleCommand(_textBox.OnDuplicate)));

				if (Document.IsCommentLineSupported || Document.IsCommentTextSupported)
				{
					subItems.Add(new MenuSeparator());

					if (Document.IsCommentLineSupported)
						subItems.Add(new MenuCommandItem("Comment Line", null, "Ctrl+/", _textBox.CommentSelectedLinesCommand));
					if (Document.IsCommentTextSupported)
						subItems.Add(new MenuCommandItem("Comment Text", null, "Ctrl+Shift+/", _textBox.CommentSelectedTextCommand));
				}

				if (Document.IsFormattingSupported)
				{
					subItems.Add(new MenuSeparator());
					subItems.Add(new MenuCommandItem("Update text format", null, "Ctrl+Alt+F", _textBox.FormattingCommand));
				}

				_menu1 = new MenuListItem("Editor", null, subItems.ToArray());
			}

			return _menu1;
		}

		public MenuListItem GenerateRefactoringMenu()
		{
			if (_menu2 == null)
			{
				List<MenuItemBase> subItems = new List<MenuItemBase>();
				subItems.Add(new MenuCommandItem("Remove empty lines", null, null, new SimpleCommand(_textBox.OnRemoveEmptyLines)));
				
				_menu2 = new MenuListItem("Refactoring", null, subItems.ToArray());
			}

			return _menu2;
		}

		public void Insert(HintItem word)
		{
			_textBox.Insert(word);
		}

		public void SaveStatus(XElement file)
		{
			file.Add(new XElement("VScrollOffset", _verticalScroll.Value.ToString(CultureInfo.InvariantCulture)));
			file.Add(new XElement("HScrollOffset", _horizontalScroll.Value.ToString(CultureInfo.InvariantCulture)));
			file.Add(new XElement("StartSelectionCol", _textBox.StartSelection.Col));
			file.Add(new XElement("StartSelectionRow", _textBox.StartSelection.Row));
			file.Add(new XElement("EndSelectionCol", _textBox.EndSelection.Col));
			file.Add(new XElement("EndSelectionRow", _textBox.EndSelection.Row));
		}

		internal void LoadState(XElement xStatusElement)
		{
			if (xStatusElement == null)
				return;

			XElement vScrollOffset = xStatusElement.Element("VScrollOffset");
			if (vScrollOffset != null)
				_verticalScroll.Value = double.Parse(vScrollOffset.Value, CultureInfo.InvariantCulture);

			XElement hScrollOffset = xStatusElement.Element("HScrollOffset");
			if (hScrollOffset != null)
				_horizontalScroll.Value = double.Parse(hScrollOffset.Value, CultureInfo.InvariantCulture);

			XElement startSelectionCol = xStatusElement.Element("StartSelectionCol");
			XElement startSelectionRow = xStatusElement.Element("StartSelectionRow");
			XElement endSelectionCol = xStatusElement.Element("EndSelectionCol");
			XElement endSelectionRow = xStatusElement.Element("EndSelectionRow");
			if (startSelectionCol != null && startSelectionRow != null && endSelectionCol != null && endSelectionRow != null)
				_textBox.SetPosition(
					new CharPosition(int.Parse(startSelectionCol.Value, CultureInfo.InvariantCulture), int.Parse(startSelectionRow.Value, CultureInfo.InvariantCulture)),
					new CharPosition(int.Parse(endSelectionCol.Value, CultureInfo.InvariantCulture), int.Parse(endSelectionRow.Value, CultureInfo.InvariantCulture)), true);
		}
	}
}
